import json

from hdmf.spec import GroupSpec, DatasetSpec, AttributeSpec, LinkSpec
from hdmf.testing import TestCase


class GroupSpecTests(TestCase):
    def setUp(self):
        self.attributes = [
            AttributeSpec('attribute1', 'my first attribute', 'text'),
            AttributeSpec('attribute2', 'my second attribute', 'text')
        ]

        self.dset1_attributes = [
            AttributeSpec('attribute3', 'my third attribute', 'text'),
            AttributeSpec('attribute4', 'my fourth attribute', 'text')
        ]

        self.dset2_attributes = [
            AttributeSpec('attribute5', 'my fifth attribute', 'text'),
            AttributeSpec('attribute6', 'my sixth attribute', 'text')
        ]

        self.datasets = [
            DatasetSpec('my first dataset',
                        'int',
                        name='dataset1',
                        attributes=self.dset1_attributes,
                        linkable=True),
            DatasetSpec('my second dataset',
                        'int',
                        name='dataset2',
                        attributes=self.dset2_attributes,
                        linkable=True,
                        data_type_def='VoltageArray')
        ]

        self.subgroups = [
            GroupSpec('A test subgroup',
                      name='subgroup1',
                      linkable=False),
            GroupSpec('A test subgroup',
                      name='subgroup2',
                      linkable=False)
        ]

    def test_constructor(self):
        spec = GroupSpec('A test group',
                         name='root_constructor',
                         groups=self.subgroups,
                         datasets=self.datasets,
                         attributes=self.attributes,
                         linkable=False)
        self.assertFalse(spec['linkable'])
        self.assertListEqual(spec['attributes'], self.attributes)
        self.assertListEqual(spec['datasets'], self.datasets)
        self.assertNotIn('data_type_def', spec)
        self.assertIs(spec, self.subgroups[0].parent)
        self.assertIs(spec, self.subgroups[1].parent)
        self.assertIs(spec, self.attributes[0].parent)
        self.assertIs(spec, self.attributes[1].parent)
        self.assertIs(spec, self.datasets[0].parent)
        self.assertIs(spec, self.datasets[1].parent)
        json.dumps(spec)

    def test_constructor_datatype(self):
        spec = GroupSpec('A test group',
                         name='root_constructor_datatype',
                         datasets=self.datasets,
                         attributes=self.attributes,
                         linkable=False,
                         data_type_def='EphysData')
        self.assertFalse(spec['linkable'])
        self.assertListEqual(spec['attributes'], self.attributes)
        self.assertListEqual(spec['datasets'], self.datasets)
        self.assertEqual(spec['data_type_def'], 'EphysData')
        self.assertIs(spec, self.attributes[0].parent)
        self.assertIs(spec, self.attributes[1].parent)
        self.assertIs(spec, self.datasets[0].parent)
        self.assertIs(spec, self.datasets[1].parent)
        self.assertEqual(spec.data_type_def, 'EphysData')
        self.assertIsNone(spec.data_type_inc)
        json.dumps(spec)

    def test_set_parent_exists(self):
        GroupSpec('A test group',
                  name='root_constructor',
                  groups=self.subgroups)
        msg = 'Cannot re-assign parent.'
        with self.assertRaisesWith(AttributeError, msg):
            self.subgroups[0].parent = self.subgroups[1]

    def test_set_dataset(self):
        spec = GroupSpec('A test group',
                         name='root_test_set_dataset',
                         linkable=False,
                         data_type_def='EphysData')
        spec.set_dataset(self.datasets[0])
        self.assertIs(spec, self.datasets[0].parent)

    def test_set_link(self):
        group = GroupSpec(
            doc='A test group',
            name='root'
        )
        link = LinkSpec(
            doc='A test link',
            target_type='LinkTarget',
            name='link_name'
        )
        group.set_link(link)
        self.assertIs(group, link.parent)
        self.assertIs(group.get_link('link_name'), link)

    def test_add_link(self):
        group = GroupSpec(
            doc='A test group',
            name='root'
        )
        group.add_link(
            'A test link',
            'LinkTarget',
            name='link_name'
        )
        self.assertIsInstance(group.get_link('link_name'), LinkSpec)

    def test_set_group(self):
        spec = GroupSpec('A test group',
                         name='root_test_set_group',
                         linkable=False,
                         data_type_def='EphysData')
        spec.set_group(self.subgroups[0])
        spec.set_group(self.subgroups[1])
        self.assertListEqual(spec['groups'], self.subgroups)
        self.assertIs(spec, self.subgroups[0].parent)
        self.assertIs(spec, self.subgroups[1].parent)
        json.dumps(spec)

    def test_add_group(self):
        group = GroupSpec(
            doc='A test group',
            name='root'
        )
        group.add_group(
            'A test group',
            name='subgroup'
        )
        self.assertIsInstance(group.get_group('subgroup'), GroupSpec)

    def test_type_extension(self):
        spec = GroupSpec('A test group',
                         name='parent_type',
                         datasets=self.datasets,
                         attributes=self.attributes,
                         linkable=False,
                         data_type_def='EphysData')
        dset1_attributes_ext = [
            AttributeSpec('dset1_extra_attribute', 'an extra attribute for the first dataset', 'text')
        ]
        ext_datasets = [
            DatasetSpec('my first dataset extension',
                        'int',
                        name='dataset1',
                        attributes=dset1_attributes_ext,
                        linkable=True),
        ]
        ext_attributes = [
            AttributeSpec('ext_extra_attribute', 'an extra attribute for the group', 'text'),
        ]
        ext = GroupSpec('A test group extension',
                        name='child_type',
                        datasets=ext_datasets,
                        attributes=ext_attributes,
                        linkable=False,
                        data_type_inc=spec,
                        data_type_def='SpikeData')
        ext_dset1 = ext.get_dataset('dataset1')
        ext_dset1_attrs = ext_dset1.attributes
        self.assertDictEqual(ext_dset1_attrs[0], dset1_attributes_ext[0])
        self.assertDictEqual(ext_dset1_attrs[1], self.dset1_attributes[0])
        self.assertDictEqual(ext_dset1_attrs[2], self.dset1_attributes[1])
        self.assertEqual(ext.data_type_def, 'SpikeData')
        self.assertEqual(ext.data_type_inc, 'EphysData')

        ext_dset2 = ext.get_dataset('dataset2')
        self.maxDiff = None
        # this will suffice for now,  assertDictEqual doesn't do deep equality checks
        self.assertEqual(str(ext_dset2), str(self.datasets[1]))
        self.assertAttributesEqual(ext_dset2, self.datasets[1])

        res_attrs = ext.attributes
        self.assertDictEqual(res_attrs[0], ext_attributes[0])
        self.assertDictEqual(res_attrs[1], self.attributes[0])
        self.assertDictEqual(res_attrs[2], self.attributes[1])

        # test that inherited specs are tracked appropriate
        for d in self.datasets:
            with self.subTest(dataset=d.name):
                self.assertTrue(ext.is_inherited_spec(d))
                self.assertFalse(spec.is_inherited_spec(d))

        json.dumps(spec)

    def assertDatasetsEqual(self, spec1, spec2):
        spec1_dsets = spec1.datasets
        spec2_dsets = spec2.datasets
        if len(spec1_dsets) != len(spec2_dsets):
            raise AssertionError('different number of AttributeSpecs')
        else:
            for i in range(len(spec1_dsets)):
                self.assertAttributesEqual(spec1_dsets[i], spec2_dsets[i])

    def assertAttributesEqual(self, spec1, spec2):
        spec1_attr = spec1.attributes
        spec2_attr = spec2.attributes
        if len(spec1_attr) != len(spec2_attr):
            raise AssertionError('different number of AttributeSpecs')
        else:
            for i in range(len(spec1_attr)):
                self.assertDictEqual(spec1_attr[i], spec2_attr[i])

    def test_add_attribute(self):
        spec = GroupSpec('A test group',
                         name='root_constructor',
                         groups=self.subgroups,
                         datasets=self.datasets,
                         linkable=False)
        for attrspec in self.attributes:
            spec.add_attribute(**attrspec)
        self.assertListEqual(spec['attributes'], self.attributes)
        self.assertListEqual(spec['datasets'], self.datasets)
        self.assertNotIn('data_type_def', spec)
        self.assertIs(spec, self.subgroups[0].parent)
        self.assertIs(spec, self.subgroups[1].parent)
        self.assertIs(spec, spec.attributes[0].parent)
        self.assertIs(spec, spec.attributes[1].parent)
        self.assertIs(spec, self.datasets[0].parent)
        self.assertIs(spec, self.datasets[1].parent)
        json.dumps(spec)

    def test_update_attribute_spec(self):
        spec = GroupSpec('A test group',
                         name='root_constructor',
                         attributes=[AttributeSpec('attribute1', 'my first attribute', 'text'),
                                     AttributeSpec('attribute2', 'my second attribute', 'text')])
        spec.set_attribute(AttributeSpec('attribute2', 'my second attribute', 'int', value=5))
        res = spec.get_attribute('attribute2')
        self.assertEqual(res.value, 5)
        self.assertEqual(res.dtype, 'int')

    def test_path(self):
        GroupSpec('A test group',
                  name='root_constructor',
                  groups=self.subgroups,
                  datasets=self.datasets,
                  attributes=self.attributes,
                  linkable=False)
        self.assertEqual(self.attributes[0].path, 'root_constructor/attribute1')
        self.assertEqual(self.datasets[0].path, 'root_constructor/dataset1')
        self.assertEqual(self.subgroups[0].path, 'root_constructor/subgroup1')

    def test_path_complicated(self):
        attribute = AttributeSpec('attribute1', 'my fifth attribute', 'text')
        dataset = DatasetSpec('my first dataset',
                              'int',
                              name='dataset1',
                              attributes=[attribute])
        subgroup = GroupSpec('A subgroup',
                             name='subgroup1',
                             datasets=[dataset])
        self.assertEqual(attribute.path, 'subgroup1/dataset1/attribute1')

        _ = GroupSpec('A test group',
                      name='root',
                      groups=[subgroup])

        self.assertEqual(attribute.path, 'root/subgroup1/dataset1/attribute1')

    def test_path_no_name(self):
        attribute = AttributeSpec('attribute1', 'my fifth attribute', 'text')
        dataset = DatasetSpec('my first dataset',
                              'int',
                              data_type_inc='DatasetType',
                              attributes=[attribute])
        subgroup = GroupSpec('A subgroup',
                             data_type_def='GroupType',
                             datasets=[dataset])
        _ = GroupSpec('A test group',
                      name='root',
                      groups=[subgroup])

        self.assertEqual(attribute.path, 'root/GroupType/DatasetType/attribute1')

    def test_data_type_property_value(self):
        """Test that the property data_type has the expected value"""
        test_cases = {
            ('Foo', 'Bar'): 'Bar',
            ('Foo', None): 'Foo',
            (None, 'Bar'): 'Bar',
            (None, None): None,
        }
        for (data_type_inc, data_type_def), data_type in test_cases.items():
            with self.subTest(data_type_inc=data_type_inc,
                              data_type_def=data_type_def, data_type=data_type):
                dataset = DatasetSpec('A dataset', 'int', name='dataset',
                                      data_type_inc=data_type_inc, data_type_def=data_type_def)
                self.assertEqual(dataset.data_type, data_type)

    def test_get_data_type_spec(self):
        expected = AttributeSpec('data_type', 'the data type of this object', 'text', value='MyType')
        self.assertDictEqual(GroupSpec.get_data_type_spec('MyType'), expected)

    def test_get_namespace_spec(self):
        expected = AttributeSpec('namespace', 'the namespace for the data type of this object', 'text', required=False)
        self.assertDictEqual(GroupSpec.get_namespace_spec(), expected)


class TestNotAllowedConfig(TestCase):

    def test_no_name_no_def_no_inc(self):
        msg = ("Cannot create Group or Dataset spec with no name without specifying 'data_type_def' "
               "and/or 'data_type_inc'.")
        with self.assertRaisesWith(ValueError, msg):
            GroupSpec('A test group')

    def test_name_with_multiple(self):
        msg = ("Cannot give specific name to something that can exist multiple times: name='MyGroup', quantity='*'")
        with self.assertRaisesWith(ValueError, msg):
            GroupSpec('A test group', name='MyGroup', quantity='*')


class TestResolveAttrs(TestCase):

    def setUp(self):
        self.def_group_spec = GroupSpec(
            doc='A test group',
            name='root',
            data_type_def='MyGroup',
            attributes=[AttributeSpec('attribute1', 'my first attribute', 'text'),
                        AttributeSpec('attribute2', 'my second attribute', 'text')]
        )
        self.inc_group_spec = GroupSpec(
            doc='A test group',
            name='root',
            data_type_inc='MyGroup',
            attributes=[AttributeSpec('attribute2', 'my second attribute', 'text', value='fixed'),
                        AttributeSpec('attribute3', 'my third attribute', 'text', value='fixed')]
        )
        self.inc_group_spec.resolve_spec(self.def_group_spec)

    def test_resolved(self):
        self.assertTupleEqual(self.inc_group_spec.attributes, (
            AttributeSpec('attribute2', 'my second attribute', 'text', value='fixed'),
            AttributeSpec('attribute3', 'my third attribute', 'text', value='fixed'),
            AttributeSpec('attribute1', 'my first attribute', 'text')
        ))

        self.assertEqual(self.inc_group_spec.get_attribute('attribute1'),
                         AttributeSpec('attribute1', 'my first attribute', 'text'))
        self.assertEqual(self.inc_group_spec.get_attribute('attribute2'),
                         AttributeSpec('attribute2', 'my second attribute', 'text', value='fixed'))
        self.assertEqual(self.inc_group_spec.get_attribute('attribute3'),
                         AttributeSpec('attribute3', 'my third attribute', 'text', value='fixed'))

        self.assertTrue(self.inc_group_spec.resolved)

    def test_is_inherited_spec(self):
        self.assertFalse(self.def_group_spec.is_inherited_spec('attribute1'))
        self.assertFalse(self.def_group_spec.is_inherited_spec('attribute2'))
        self.assertTrue(self.inc_group_spec.is_inherited_spec(
            AttributeSpec('attribute1', 'my first attribute', 'text')
        ))
        self.assertTrue(self.inc_group_spec.is_inherited_spec('attribute1'))
        self.assertTrue(self.inc_group_spec.is_inherited_spec('attribute2'))
        self.assertFalse(self.inc_group_spec.is_inherited_spec('attribute3'))
        self.assertFalse(self.inc_group_spec.is_inherited_spec('attribute4'))

    def test_is_overridden_spec(self):
        self.assertFalse(self.def_group_spec.is_overridden_spec('attribute1'))
        self.assertFalse(self.def_group_spec.is_overridden_spec('attribute2'))
        self.assertFalse(self.inc_group_spec.is_overridden_spec(
            AttributeSpec('attribute1', 'my first attribute', 'text')
        ))
        self.assertFalse(self.inc_group_spec.is_overridden_spec('attribute1'))
        self.assertTrue(self.inc_group_spec.is_overridden_spec('attribute2'))
        self.assertFalse(self.inc_group_spec.is_overridden_spec('attribute3'))
        self.assertFalse(self.inc_group_spec.is_overridden_spec('attribute4'))

    def test_is_inherited_attribute(self):
        self.assertFalse(self.def_group_spec.is_inherited_attribute('attribute1'))
        self.assertFalse(self.def_group_spec.is_inherited_attribute('attribute2'))
        self.assertTrue(self.inc_group_spec.is_inherited_attribute('attribute1'))
        self.assertTrue(self.inc_group_spec.is_inherited_attribute('attribute2'))
        self.assertFalse(self.inc_group_spec.is_inherited_attribute('attribute3'))
        with self.assertRaisesWith(ValueError, "Attribute 'attribute4' not found"):
            self.inc_group_spec.is_inherited_attribute('attribute4')

    def test_is_overridden_attribute(self):
        self.assertFalse(self.def_group_spec.is_overridden_attribute('attribute1'))
        self.assertFalse(self.def_group_spec.is_overridden_attribute('attribute2'))
        self.assertFalse(self.inc_group_spec.is_overridden_attribute('attribute1'))
        self.assertTrue(self.inc_group_spec.is_overridden_attribute('attribute2'))
        self.assertFalse(self.inc_group_spec.is_overridden_attribute('attribute3'))
        with self.assertRaisesWith(ValueError, "Attribute 'attribute4' not found"):
            self.inc_group_spec.is_overridden_attribute('attribute4')


class GroupSpecWithLinksTest(TestCase):

    def test_constructor(self):
        link0 = LinkSpec(doc='Link 0', target_type='TargetType0')
        link1 = LinkSpec(doc='Link 1', target_type='TargetType1')
        links = [link0, link1]
        spec = GroupSpec(
            doc='A test group',
            name='root',
            links=links
        )
        self.assertIs(spec, links[0].parent)
        self.assertIs(spec, links[1].parent)
        json.dumps(spec)

    def test_extension_no_overwrite(self):
        link0 = LinkSpec(doc='Link 0', target_type='TargetType0')  # test unnamed
        link1 = LinkSpec(doc='Link 1', target_type='TargetType1', name='MyType1')  # test named
        link2 = LinkSpec(doc='Link 2', target_type='TargetType2', quantity='*')  # test named, multiple
        links = [link0, link1, link2]
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=links,
            data_type_def='ParentType'
        )
        child_spec = GroupSpec(
            doc='A test group',
            name='child',
            data_type_inc=parent_spec,
            data_type_def='ChildType'
        )

        for link in links:
            with self.subTest(link_target_type=link.target_type):
                self.assertTrue(child_spec.is_inherited_spec(link))
                self.assertFalse(child_spec.is_overridden_spec(link))

    def test_extension_overwrite(self):
        link0 = LinkSpec(doc='Link 0', target_type='TargetType0', name='MyType0')
        link1 = LinkSpec(doc='Link 1', target_type='TargetType1', name='MyType1')
        # NOTE overwriting unnamed LinkSpec is not allowed
        # NOTE overwriting spec with quantity that could be >1 is not allowed
        links = [link0, link1]
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=links,
            data_type_def='ParentType'
        )

        link0_overwrite = LinkSpec(doc='New link 0', target_type='TargetType0', name='MyType0')
        link1_overwrite = LinkSpec(doc='New link 1', target_type='TargetType1Child', name='MyType1')
        overwritten_links = [link0_overwrite, link1_overwrite]
        child_spec = GroupSpec(
            doc='A test group',
            name='child',
            links=overwritten_links,
            data_type_inc=parent_spec,
            data_type_def='ChildType'
        )

        for link in overwritten_links:
            with self.subTest(link_target_type=link.target_type):
                self.assertTrue(child_spec.is_inherited_spec(link))
                self.assertTrue(child_spec.is_overridden_spec(link))


class SpecWithDupsTest(TestCase):

    def test_two_unnamed_group_same_type(self):
        """Test creating a group contains multiple unnamed groups with type X."""
        child0 = GroupSpec(doc='Group 0', data_type_inc='Type0')
        child1 = GroupSpec(doc='Group 1', data_type_inc='Type0')
        msg = "Cannot have multiple groups/datasets with the same data type without specifying name"
        with self.assertRaisesWith(ValueError, msg):
            GroupSpec(
                doc='A test group',
                name='parent',
                groups=[child0, child1],
                data_type_def='ParentType'
            )

    def test_named_unnamed_group_with_def_same_type(self):
        """Test get_data_type when a group contains both a named and unnamed group with type X."""
        child0 = GroupSpec(doc='Group 0', data_type_def='Type0', name='type0')
        child1 = GroupSpec(doc='Group 1', data_type_inc='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            groups=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child1)

    def test_named_unnamed_group_same_type(self):
        """Test get_data_type when a group contains both a named and unnamed group with type X."""
        child0 = GroupSpec(doc='Group 0', data_type_inc='Type0', name='type0')
        child1 = GroupSpec(doc='Group 1', data_type_inc='Type0', name='type1')
        child2 = GroupSpec(doc='Group 2', data_type_inc='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            groups=[child0, child1, child2],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child2)

    def test_unnamed_named_group_same_type(self):
        """Test get_data_type when a group contains both an unnamed and named group with type X."""
        child0 = GroupSpec(doc='Group 0', data_type_inc='Type0')
        child1 = GroupSpec(doc='Group 1', data_type_inc='Type0', name='type1')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            groups=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child0)

    def test_two_named_group_same_type(self):
        """Test get_data_type when a group contains multiple named groups with type X."""
        child0 = GroupSpec(doc='Group 0', data_type_inc='Type0', name='group0')
        child1 = GroupSpec(doc='Group 1', data_type_inc='Type0', name='group1')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            groups=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertEqual(parent_spec.get_data_type('Type0'), [child0, child1])

    def test_two_unnamed_datasets_same_type(self):
        """Test creating a group contains multiple unnamed datasets with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_inc='Type0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0')
        msg = "Cannot have multiple groups/datasets with the same data type without specifying name"
        with self.assertRaisesWith(ValueError, msg):
            GroupSpec(
                doc='A test group',
                name='parent',
                datasets=[child0, child1],
                data_type_def='ParentType'
            )

    def test_named_unnamed_dataset_with_def_same_type(self):
        """Test get_data_type when a group contains both a named and unnamed dataset with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_def='Type0', name='type0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child1)

    def test_named_unnamed_dataset_same_type(self):
        """Test get_data_type when a group contains both a named and unnamed dataset with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_inc='Type0', name='type0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child1)

    def test_two_named_unnamed_dataset_same_type(self):
        """Test get_data_type when a group contains two named and one unnamed dataset with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_inc='Type0', name='type0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0', name='type1')
        child2 = DatasetSpec(doc='Group 2', data_type_inc='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child0, child1, child2],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child2)

    def test_unnamed_named_dataset_same_type(self):
        """Test get_data_type when a group contains both an unnamed and named dataset with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_inc='Type0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0', name='type1')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child0)

    def test_two_named_datasets_same_type(self):
        """Test get_data_type when a group contains multiple named datasets with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_inc='Type0', name='group0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0', name='group1')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertEqual(parent_spec.get_data_type('Type0'), [child0, child1])

    def test_three_named_datasets_same_type(self):
        """Test get_target_type when a group contains three named links with type X."""
        child0 = DatasetSpec(doc='Group 0', data_type_inc='Type0', name='group0')
        child1 = DatasetSpec(doc='Group 1', data_type_inc='Type0', name='group1')
        child2 = DatasetSpec(doc='Group 2', data_type_inc='Type0', name='group2')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child0, child1, child2],
            data_type_def='ParentType'
        )
        self.assertEqual(parent_spec.get_data_type('Type0'), [child0, child1, child2])

    def test_two_unnamed_links_same_type(self):
        """Test creating a group contains multiple unnamed links with type X."""
        child0 = LinkSpec(doc='Group 0', target_type='Type0')
        child1 = LinkSpec(doc='Group 1', target_type='Type0')
        msg = "Cannot have multiple links with the same target type without specifying name"
        with self.assertRaisesWith(ValueError, msg):
            GroupSpec(
                doc='A test group',
                name='parent',
                links=[child0, child1],
                data_type_def='ParentType'
            )

    def test_named_unnamed_link_same_type(self):
        """Test get_target_type when a group contains both a named and unnamed link with type X."""
        child0 = LinkSpec(doc='Group 0', target_type='Type0', name='type0')
        child1 = LinkSpec(doc='Group 1', target_type='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_target_type('Type0'), child1)

    def test_two_named_unnamed_link_same_type(self):
        """Test get_target_type when a group contains two named and one unnamed link with type X."""
        child0 = LinkSpec(doc='Group 0', target_type='Type0', name='type0')
        child1 = LinkSpec(doc='Group 1', target_type='Type0', name='type1')
        child2 = LinkSpec(doc='Group 2', target_type='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=[child0, child1, child2],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_target_type('Type0'), child2)

    def test_unnamed_named_link_same_type(self):
        """Test get_target_type when a group contains both an unnamed and named link with type X."""
        child0 = LinkSpec(doc='Group 0', target_type='Type0')
        child1 = LinkSpec(doc='Group 1', target_type='Type0', name='type1')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_target_type('Type0'), child0)

    def test_two_named_links_same_type(self):
        """Test get_target_type when a group contains multiple named links with type X."""
        child0 = LinkSpec(doc='Group 0', target_type='Type0', name='group0')
        child1 = LinkSpec(doc='Group 1', target_type='Type0', name='group1')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=[child0, child1],
            data_type_def='ParentType'
        )
        self.assertEqual(parent_spec.get_target_type('Type0'), [child0, child1])

    def test_three_named_links_same_type(self):
        """Test get_target_type when a group contains three named links with type X."""
        child0 = LinkSpec(doc='Group 0', target_type='Type0', name='type0')
        child1 = LinkSpec(doc='Group 1', target_type='Type0', name='type1')
        child2 = LinkSpec(doc='Group 2', target_type='Type0', name='type2')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            links=[child0, child1, child2],
            data_type_def='ParentType'
        )
        self.assertEqual(parent_spec.get_target_type('Type0'), [child0, child1, child2])


class SpecWithGroupsLinksTest(TestCase):

    def test_unnamed_group_link_same_type(self):
        child = GroupSpec(doc='Group 0', data_type_inc='Type0')
        link = LinkSpec(doc='Link 0', target_type='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            groups=[child],
            links=[link],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child)
        self.assertIs(parent_spec.get_target_type('Type0'), link)

    def test_unnamed_dataset_link_same_type(self):
        child = DatasetSpec(doc='Dataset 0', data_type_inc='Type0')
        link = LinkSpec(doc='Link 0', target_type='Type0')
        parent_spec = GroupSpec(
            doc='A test group',
            name='parent',
            datasets=[child],
            links=[link],
            data_type_def='ParentType'
        )
        self.assertIs(parent_spec.get_data_type('Type0'), child)
        self.assertIs(parent_spec.get_target_type('Type0'), link)



import pandas as pd

from hdmf.container import Table, Row, RowGetter
from hdmf.testing import TestCase


class TestTable(TestCase):

    @classmethod
    def get_table_class(cls):
        class MyTable(Table):

            __defaultname__ = 'my_table'

            __columns__ = [
                {'name': 'col1', 'type': str, 'help': 'a string column'},
                {'name': 'col2', 'type': int, 'help': 'an integer column'},
            ]
        return MyTable

    def test_init(self):
        MyTable = TestTable.get_table_class()
        table = MyTable('test_table')
        self.assertTrue(hasattr(table, '__colidx__'))
        self.assertEqual(table.__colidx__, {'col1': 0, 'col2': 1})

    def test_add_row_getitem(self):
        MyTable = TestTable.get_table_class()
        table = MyTable('test_table')
        table.add_row(col1='foo', col2=100)
        table.add_row(col1='bar', col2=200)
        row1 = table[0]
        row2 = table[1]
        self.assertEqual(row1, ('foo', 100))
        self.assertEqual(row2, ('bar', 200))

    def test_to_dataframe(self):
        MyTable = TestTable.get_table_class()
        table = MyTable('test_table')
        table.add_row(col1='foo', col2=100)
        table.add_row(col1='bar', col2=200)

        df = table.to_dataframe()
        exp = pd.DataFrame(data=[{'col1': 'foo', 'col2': 100}, {'col1': 'bar', 'col2': 200}])
        pd.testing.assert_frame_equal(df, exp)

    def test_from_dataframe(self):
        MyTable = TestTable.get_table_class()
        exp = pd.DataFrame(data=[{'col1': 'foo', 'col2': 100}, {'col1': 'bar', 'col2': 200}])
        table = MyTable.from_dataframe(exp)
        row1 = table[0]
        row2 = table[1]
        self.assertEqual(row1, ('foo', 100))
        self.assertEqual(row2, ('bar', 200))


class TestRow(TestCase):

    def setUp(self):
        self.MyTable = TestTable.get_table_class()

        class MyRow(Row):
            __table__ = self.MyTable

        self.MyRow = MyRow

        self.table = self.MyTable('test_table')

    def test_row_no_table(self):
        with self.assertRaisesRegex(ValueError, '__table__ must be set if sub-classing Row'):
            class MyRow(Row):
                pass

    def test_table_init(self):
        MyTable = TestTable.get_table_class()
        table = MyTable('test_table')
        self.assertFalse(hasattr(table, 'row'))

        table_w_row = self.MyTable('test_table')
        self.assertTrue(hasattr(table_w_row, 'row'))
        self.assertIsInstance(table_w_row.row, RowGetter)
        self.assertIs(table_w_row.row.table, table_w_row)

    def test_init(self):
        row1 = self.MyRow(col1='foo', col2=100, table=self.table)

        # make sure Row object set up properly
        self.assertEqual(row1.idx, 0)
        self.assertEqual(row1.col1, 'foo')
        self.assertEqual(row1.col2, 100)

        # make sure Row object is stored in Table peroperly
        tmp_row1 = self.table.row[0]
        self.assertEqual(tmp_row1, row1)

    def test_add_row_getitem(self):
        self.table.add_row(col1='foo', col2=100)
        self.table.add_row(col1='bar', col2=200)

        row1 = self.table.row[0]
        self.assertIsInstance(row1, self.MyRow)
        self.assertEqual(row1.idx, 0)
        self.assertEqual(row1.col1, 'foo')
        self.assertEqual(row1.col2, 100)

        row2 = self.table.row[1]
        self.assertIsInstance(row2, self.MyRow)
        self.assertEqual(row2.idx, 1)
        self.assertEqual(row2.col1, 'bar')
        self.assertEqual(row2.col2, 200)

        # test memoization
        row3 = self.table.row[0]
        self.assertIs(row3, row1)

    def test_todict(self):
        row1 = self.MyRow(col1='foo', col2=100, table=self.table)
        self.assertEqual(row1.todict(), {'col1': 'foo', 'col2': 100})

    def test___str__(self):
        row1 = self.MyRow(col1='foo', col2=100, table=self.table)
        row1_str = str(row1)
        expected_str = "Row(0, test_table) = {'col1': 'foo', 'col2': 100}"
        self.assertEqual(row1_str, expected_str)



from hdmf import Container, Data
from hdmf.build import (BuildManager, TypeMap, GroupBuilder, DatasetBuilder, LinkBuilder, ObjectMapper,
                        MissingRequiredBuildWarning, IncorrectQuantityBuildWarning)
from hdmf.spec import GroupSpec, DatasetSpec, LinkSpec, SpecCatalog, SpecNamespace, NamespaceCatalog
from hdmf.spec.spec import ZERO_OR_MANY, ONE_OR_MANY, ZERO_OR_ONE, DEF_QUANTITY
from hdmf.testing import TestCase
from hdmf.utils import docval, getargs

from tests.unit.utils import CORE_NAMESPACE


##########################
# test all crosses:
# {
#   untyped, named group with data-type-included groups / data-type-included datasets / links
#   nested, type definition
#   included groups / included datasets / links
# }
# x
# group/dataset/link with quantity {'*', '+', 1, 2, '?'}
# x
# builder with 2, 1, or 0 instances of the type, or 0 instances of the type with some instances of a mismatched type


class SimpleFoo(Container):
    pass


class NotSimpleFoo(Container):
    pass


class SimpleQux(Data):
    pass


class NotSimpleQux(Data):
    pass


class SimpleBucket(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this SimpleBucket'},
            {'name': 'foos', 'type': list, 'doc': 'the SimpleFoo objects', 'default': list()},
            {'name': 'quxs', 'type': list, 'doc': 'the SimpleQux objects', 'default': list()},
            {'name': 'links', 'type': list, 'doc': 'another way to store SimpleFoo objects', 'default': list()})
    def __init__(self, **kwargs):
        name, foos, quxs, links = getargs('name', 'foos', 'quxs', 'links', kwargs)
        super().__init__(name=name)
        # note: collections of groups are unordered in HDF5, so make these dictionaries for keyed access
        self.foos = {f.name: f for f in foos}
        for f in foos:
            f.parent = self
        self.quxs = {q.name: q for q in quxs}
        for q in quxs:
            q.parent = self
        self.links = {i.name: i for i in links}
        for i in links:
            i.parent = self


class BasicBucket(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this BasicBucket'},
            {'name': 'untyped_dataset', 'type': 'scalar_data',
             'doc': 'a scalar dataset within this BasicBucket', 'default': None},
            {'name': 'untyped_array_dataset', 'type': ('data', 'array_data'),
             'doc': 'an array dataset within this BasicBucket', 'default': None},)
    def __init__(self, **kwargs):
        name, untyped_dataset, untyped_array_dataset = getargs('name', 'untyped_dataset', 'untyped_array_dataset',
                                                               kwargs)
        super().__init__(name=name)
        self.untyped_dataset = untyped_dataset
        self.untyped_array_dataset = untyped_array_dataset


class BuildQuantityMixin:
    """Base test class mixin to set up the BuildManager."""

    def setUpManager(self, specs):
        spec_catalog = SpecCatalog()
        schema_file = 'test.yaml'
        for s in specs:
            spec_catalog.register_spec(s, schema_file)
        namespace = SpecNamespace(
            doc='a test namespace',
            name=CORE_NAMESPACE,
            schema=[{'source': schema_file}],
            version='0.1.0',
            catalog=spec_catalog
        )
        namespace_catalog = NamespaceCatalog()
        namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
        type_map = TypeMap(namespace_catalog)
        type_map.register_container_type(CORE_NAMESPACE, 'SimpleFoo', SimpleFoo)
        type_map.register_container_type(CORE_NAMESPACE, 'NotSimpleFoo', NotSimpleFoo)
        type_map.register_container_type(CORE_NAMESPACE, 'SimpleQux', SimpleQux)
        type_map.register_container_type(CORE_NAMESPACE, 'NotSimpleQux', NotSimpleQux)
        type_map.register_container_type(CORE_NAMESPACE, 'SimpleBucket', SimpleBucket)
        type_map.register_map(SimpleBucket, self.setUpBucketMapper())
        self.manager = BuildManager(type_map)

    def _create_builder(self, container):
        """Helper function to get a basic builder for a container with no subgroups/datasets/links."""
        if isinstance(container, Container):
            ret = GroupBuilder(
                name=container.name,
                attributes={'namespace': container.namespace,
                            'data_type': container.data_type,
                            'object_id': container.object_id}
            )
        else:
            ret = DatasetBuilder(
                name=container.name,
                data=container.data,
                attributes={'namespace': container.namespace,
                            'data_type': container.data_type,
                            'object_id': container.object_id}
            )
        return ret


class TypeIncUntypedGroupMixin:

    def create_specs(self, quantity):
        # Type SimpleBucket contains:
        # - an untyped group "foo_holder" which contains [quantity] groups of data_type_inc SimpleFoo
        # - an untyped group "qux_holder" which contains [quantity] datasets of data_type_inc SimpleQux
        # - an untyped group "link_holder" which contains [quantity] links of target_type SimpleFoo
        foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='SimpleFoo',
        )
        not_foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleFoo',
        )
        qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='SimpleQux',
        )
        not_qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleQux',
        )
        foo_inc_spec = GroupSpec(
            doc='the SimpleFoos in this bucket',
            data_type_inc='SimpleFoo',
            quantity=quantity
        )
        foo_holder_spec = GroupSpec(
            doc='An untyped subgroup for SimpleFoos',
            name='foo_holder',
            groups=[foo_inc_spec]
        )
        qux_inc_spec = DatasetSpec(
            doc='the SimpleQuxs in this bucket',
            data_type_inc='SimpleQux',
            quantity=quantity
        )
        qux_holder_spec = GroupSpec(
            doc='An untyped subgroup for SimpleQuxs',
            name='qux_holder',
            datasets=[qux_inc_spec]
        )
        foo_link_spec = LinkSpec(
            doc='the links in this bucket',
            target_type='SimpleFoo',
            quantity=quantity
        )
        link_holder_spec = GroupSpec(
            doc='An untyped subgroup for links',
            name='link_holder',
            links=[foo_link_spec]
        )
        bucket_spec = GroupSpec(
            doc='A test group specification for a data type containing data type',
            name="test_bucket",
            data_type_def='SimpleBucket',
            groups=[foo_holder_spec, qux_holder_spec, link_holder_spec]
        )
        return [foo_spec, not_foo_spec, qux_spec, not_qux_spec, bucket_spec]

    def setUpBucketMapper(self):
        class BucketMapper(ObjectMapper):
            def __init__(self, spec):
                super().__init__(spec)
                self.unmap(spec.get_group('foo_holder'))
                self.map_spec('foos', spec.get_group('foo_holder').get_data_type('SimpleFoo'))
                self.unmap(spec.get_group('qux_holder'))
                self.map_spec('quxs', spec.get_group('qux_holder').get_data_type('SimpleQux'))
                self.unmap(spec.get_group('link_holder'))
                self.map_spec('links', spec.get_group('link_holder').links[0])

        return BucketMapper

    def get_two_bucket_test(self):
        foos = [SimpleFoo('my_foo1'), SimpleFoo('my_foo2')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3]), SimpleQux('my_qux2', data=[4, 5, 6])]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=foos
        )
        foo1_builder = self._create_builder(bucket.foos['my_foo1'])
        foo2_builder = self._create_builder(bucket.foos['my_foo2'])
        qux1_builder = self._create_builder(bucket.quxs['my_qux1'])
        qux2_builder = self._create_builder(bucket.quxs['my_qux2'])
        foo_holder_builder = GroupBuilder(
            name='foo_holder',
            groups={'my_foo1': foo1_builder,
                    'my_foo2': foo2_builder}
        )
        qux_holder_builder = GroupBuilder(
            name='qux_holder',
            datasets={'my_qux1': qux1_builder,
                      'my_qux2': qux2_builder}
        )
        foo1_link_builder = LinkBuilder(builder=foo1_builder)
        foo2_link_builder = LinkBuilder(builder=foo2_builder)
        link_holder_builder = GroupBuilder(
            name='link_holder',
            links={'my_foo1': foo1_link_builder,
                   'my_foo2': foo2_link_builder}
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'foos': foo_holder_builder,
                    'quxs': qux_holder_builder,
                    'links': link_holder_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_one_bucket_test(self):
        foos = [SimpleFoo('my_foo1')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3])]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=foos
        )
        foo1_builder = GroupBuilder(
            name='my_foo1',
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleFoo',
                        'object_id': bucket.foos['my_foo1'].object_id}
        )
        foo_holder_builder = GroupBuilder(
            name='foo_holder',
            groups={'my_foo1': foo1_builder}
        )
        qux1_builder = DatasetBuilder(
            name='my_qux1',
            data=[1, 2, 3],
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleQux',
                        'object_id': bucket.quxs['my_qux1'].object_id}
        )
        qux_holder_builder = GroupBuilder(
            name='qux_holder',
            datasets={'my_qux1': qux1_builder}
        )
        foo1_link_builder = LinkBuilder(builder=foo1_builder)
        link_holder_builder = GroupBuilder(
            name='link_holder',
            links={'my_foo1': foo1_link_builder}
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'foos': foo_holder_builder,
                    'quxs': qux_holder_builder,
                    'links': link_holder_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_zero_bucket_test(self):
        bucket = SimpleBucket(
            name='test_bucket'
        )
        foo_holder_builder = GroupBuilder(
            name='foo_holder',
            groups={}
        )
        qux_holder_builder = GroupBuilder(
            name='qux_holder',
            datasets={}
        )
        link_holder_builder = GroupBuilder(
            name='link_holder',
            links={}
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'foos': foo_holder_builder,
                    'quxs': qux_holder_builder,
                    'links': link_holder_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_mismatch_bucket_test(self):
        foos = [NotSimpleFoo('my_foo1'), NotSimpleFoo('my_foo2')]
        quxs = [NotSimpleQux('my_qux1', data=[1, 2, 3]), NotSimpleQux('my_qux2', data=[4, 5, 6])]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=foos
        )
        foo_holder_builder = GroupBuilder(
            name='foo_holder',
            groups={}
        )
        qux_holder_builder = GroupBuilder(
            name='qux_holder',
            datasets={}
        )
        link_holder_builder = GroupBuilder(
            name='link_holder',
            links={}
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'foos': foo_holder_builder,
                    'quxs': qux_holder_builder,
                    'links': link_holder_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder


class TypeDefMixin:

    def create_specs(self, quantity):
        # Type SimpleBucket contains:
        # - contains [quantity] groups of data_type_def SimpleFoo
        # - contains [quantity] datasets of data_type_def SimpleQux
        # NOTE: links do not have data_type_def, so leave them out of these tests
        # NOTE: nested type definitions are strongly discouraged now
        foo_spec = GroupSpec(
            doc='the SimpleFoos in this bucket',
            data_type_def='SimpleFoo',
            quantity=quantity
        )
        qux_spec = DatasetSpec(
            doc='the SimpleQuxs in this bucket',
            data_type_def='SimpleQux',
            quantity=quantity
        )
        not_foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleFoo',
        )
        not_qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleQux',
        )
        bucket_spec = GroupSpec(
            doc='A test group specification for a data type containing data type',
            name="test_bucket",
            data_type_def='SimpleBucket',
            groups=[foo_spec],
            datasets=[qux_spec]
        )
        return [foo_spec, not_foo_spec, qux_spec, not_qux_spec, bucket_spec]

    def setUpBucketMapper(self):
        class BucketMapper(ObjectMapper):
            def __init__(self, spec):
                super().__init__(spec)
                self.map_spec('foos', spec.get_data_type('SimpleFoo'))
                self.map_spec('quxs', spec.get_data_type('SimpleQux'))

        return BucketMapper

    def get_two_bucket_test(self):
        foos = [SimpleFoo('my_foo1'), SimpleFoo('my_foo2')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3]), SimpleQux('my_qux2', data=[4, 5, 6])]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
        )
        foo1_builder = self._create_builder(bucket.foos['my_foo1'])
        foo2_builder = self._create_builder(bucket.foos['my_foo2'])
        qux1_builder = self._create_builder(bucket.quxs['my_qux1'])
        qux2_builder = self._create_builder(bucket.quxs['my_qux2'])
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'my_foo1': foo1_builder,
                    'my_foo2': foo2_builder},
            datasets={'my_qux1': qux1_builder,
                      'my_qux2': qux2_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_one_bucket_test(self):
        foos = [SimpleFoo('my_foo1')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3])]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
        )
        foo1_builder = self._create_builder(bucket.foos['my_foo1'])
        qux1_builder = self._create_builder(bucket.quxs['my_qux1'])
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'my_foo1': foo1_builder},
            datasets={'my_qux1': qux1_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_zero_bucket_test(self):
        bucket = SimpleBucket(
            name='test_bucket'
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_mismatch_bucket_test(self):
        foos = [NotSimpleFoo('my_foo1'), NotSimpleFoo('my_foo2')]
        quxs = [NotSimpleQux('my_qux1', data=[1, 2, 3]), NotSimpleQux('my_qux2', data=[4, 5, 6])]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder


class TypeIncMixin:

    def create_specs(self, quantity):
        # Type SimpleBucket contains:
        # - [quantity] groups of data_type_inc SimpleFoo
        # - [quantity] datasets of data_type_inc SimpleQux
        # - [quantity] links of target_type SimpleFoo
        foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='SimpleFoo',
        )
        not_foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleFoo',
        )
        qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='SimpleQux',
        )
        not_qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleQux',
        )
        foo_inc_spec = GroupSpec(
            doc='the SimpleFoos in this bucket',
            data_type_inc='SimpleFoo',
            quantity=quantity
        )
        qux_inc_spec = DatasetSpec(
            doc='the SimpleQuxs in this bucket',
            data_type_inc='SimpleQux',
            quantity=quantity
        )
        foo_link_spec = LinkSpec(
            doc='the links in this bucket',
            target_type='SimpleFoo',
            quantity=quantity
        )
        bucket_spec = GroupSpec(
            doc='A test group specification for a data type containing data type',
            name="test_bucket",
            data_type_def='SimpleBucket',
            groups=[foo_inc_spec],
            datasets=[qux_inc_spec],
            links=[foo_link_spec]
        )
        return [foo_spec, not_foo_spec, qux_spec, not_qux_spec, bucket_spec]

    def setUpBucketMapper(self):
        class BucketMapper(ObjectMapper):
            def __init__(self, spec):
                super().__init__(spec)
                self.map_spec('foos', spec.get_data_type('SimpleFoo'))
                self.map_spec('quxs', spec.get_data_type('SimpleQux'))
                self.map_spec('links', spec.links[0])

        return BucketMapper

    def get_two_bucket_test(self):
        foos = [SimpleFoo('my_foo1'), SimpleFoo('my_foo2')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3]), SimpleQux('my_qux2', data=[4, 5, 6])]
        # NOTE: unlike in the other tests, links cannot map to the same foos in bucket because of a name clash
        links = [SimpleFoo('my_foo3'), SimpleFoo('my_foo4')]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=links
        )
        foo1_builder = self._create_builder(bucket.foos['my_foo1'])
        foo2_builder = self._create_builder(bucket.foos['my_foo2'])
        foo3_builder = self._create_builder(bucket.links['my_foo3'])
        foo4_builder = self._create_builder(bucket.links['my_foo4'])
        qux1_builder = self._create_builder(bucket.quxs['my_qux1'])
        qux2_builder = self._create_builder(bucket.quxs['my_qux2'])
        foo3_link_builder = LinkBuilder(builder=foo3_builder)
        foo4_link_builder = LinkBuilder(builder=foo4_builder)
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'my_foo1': foo1_builder,
                    'my_foo2': foo2_builder},
            datasets={'my_qux1': qux1_builder,
                      'my_qux2': qux2_builder},
            links={'my_foo3': foo3_link_builder,
                   'my_foo4': foo4_link_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_one_bucket_test(self):
        foos = [SimpleFoo('my_foo1')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3])]
        # NOTE: unlike in the other tests, links cannot map to the same foos in bucket because of a name clash
        links = [SimpleFoo('my_foo3')]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=links
        )
        foo1_builder = self._create_builder(bucket.foos['my_foo1'])
        foo3_builder = self._create_builder(bucket.links['my_foo3'])
        qux1_builder = self._create_builder(bucket.quxs['my_qux1'])
        foo3_link_builder = LinkBuilder(builder=foo3_builder)
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'my_foo1': foo1_builder},
            datasets={'my_qux1': qux1_builder},
            links={'my_foo1': foo3_link_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_zero_bucket_test(self):
        bucket = SimpleBucket(
            name='test_bucket'
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def get_mismatch_bucket_test(self):
        foos = [NotSimpleFoo('my_foo1'), NotSimpleFoo('my_foo2')]
        quxs = [NotSimpleQux('my_qux1', data=[1, 2, 3]), NotSimpleQux('my_qux2', data=[4, 5, 6])]
        links = [NotSimpleFoo('my_foo1'), NotSimpleFoo('my_foo2')]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=links
        )
        bucket_builder = GroupBuilder(
            name='test_bucket',
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder


class ZeroOrManyMixin:

    def setUp(self):
        specs = self.create_specs(ZERO_OR_MANY)
        self.setUpManager(specs)

    def test_build_two(self):
        """Test building a container which contains multiple containers as the spec allows."""
        bucket, bucket_builder = self.get_two_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_one(self):
        """Test building a container which contains one container as the spec allows."""
        bucket, bucket_builder = self.get_one_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_zero(self):
        """Test building a container which contains no containers as the spec allows."""
        bucket, bucket_builder = self.get_zero_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_mismatch(self):
        """Test building a container which contains no containers that match the spec as the spec allows."""
        bucket, bucket_builder = self.get_mismatch_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


class OneOrManyMixin:

    def setUp(self):
        specs = self.create_specs(ONE_OR_MANY)
        self.setUpManager(specs)

    def test_build_two(self):
        """Test building a container which contains multiple containers as the spec allows."""
        bucket, bucket_builder = self.get_two_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_one(self):
        """Test building a container which contains one container as the spec allows."""
        bucket, bucket_builder = self.get_one_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_zero(self):
        """Test building a container which contains no containers as the spec allows."""
        bucket, bucket_builder = self.get_zero_bucket_test()
        msg = r"SimpleBucket 'test_bucket' is missing required value for attribute '.*'\."
        with self.assertWarnsRegex(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_mismatch(self):
        """Test building a container which contains no containers that match the spec as the spec allows."""
        bucket, bucket_builder = self.get_mismatch_bucket_test()
        msg = r"SimpleBucket 'test_bucket' is missing required value for attribute '.*'\."
        with self.assertWarnsRegex(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


class OneMixin:

    def setUp(self):
        specs = self.create_specs(DEF_QUANTITY)
        self.setUpManager(specs)

    def test_build_two(self):
        """Test building a container which contains multiple containers as the spec allows."""
        bucket, bucket_builder = self.get_two_bucket_test()
        msg = r"SimpleBucket 'test_bucket' has 2 values for attribute '.*' but spec allows 1\."
        with self.assertWarnsRegex(IncorrectQuantityBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_one(self):
        """Test building a container which contains one container as the spec allows."""
        bucket, bucket_builder = self.get_one_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_zero(self):
        """Test building a container which contains no containers as the spec allows."""
        bucket, bucket_builder = self.get_zero_bucket_test()
        msg = r"SimpleBucket 'test_bucket' is missing required value for attribute '.*'\."
        with self.assertWarnsRegex(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_mismatch(self):
        """Test building a container which contains no containers that match the spec as the spec allows."""
        bucket, bucket_builder = self.get_mismatch_bucket_test()
        msg = r"SimpleBucket 'test_bucket' is missing required value for attribute '.*'\."
        with self.assertWarnsRegex(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


class TwoMixin:

    def setUp(self):
        specs = self.create_specs(2)
        self.setUpManager(specs)

    def test_build_two(self):
        """Test building a container which contains multiple containers as the spec allows."""
        bucket, bucket_builder = self.get_two_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_one(self):
        """Test building a container which contains one container as the spec allows."""
        bucket, bucket_builder = self.get_one_bucket_test()
        msg = r"SimpleBucket 'test_bucket' has 1 values for attribute '.*' but spec allows 2\."
        with self.assertWarnsRegex(IncorrectQuantityBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_zero(self):
        """Test building a container which contains no containers as the spec allows."""
        bucket, bucket_builder = self.get_zero_bucket_test()
        msg = r"SimpleBucket 'test_bucket' is missing required value for attribute '.*'\."
        with self.assertWarnsRegex(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_mismatch(self):
        """Test building a container which contains no containers that match the spec as the spec allows."""
        bucket, bucket_builder = self.get_mismatch_bucket_test()
        msg = r"SimpleBucket 'test_bucket' is missing required value for attribute '.*'\."
        with self.assertWarnsRegex(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


class ZeroOrOneMixin:

    def setUp(self):
        specs = self.create_specs(ZERO_OR_ONE)
        self.setUpManager(specs)

    def test_build_two(self):
        """Test building a container which contains multiple containers as the spec allows."""
        bucket, bucket_builder = self.get_two_bucket_test()
        msg = r"SimpleBucket 'test_bucket' has 2 values for attribute '.*' but spec allows '\?'\."
        with self.assertWarnsRegex(IncorrectQuantityBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_one(self):
        """Test building a container which contains one container as the spec allows."""
        bucket, bucket_builder = self.get_one_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_zero(self):
        """Test building a container which contains no containers as the spec allows."""
        bucket, bucket_builder = self.get_zero_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_mismatch(self):
        """Test building a container which contains no containers that match the spec as the spec allows."""
        bucket, bucket_builder = self.get_mismatch_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


# Untyped group with included groups / included datasets / links with quantity {'*', '+', 1, 2 '?'}

class TestBuildZeroOrManyTypeIncUntypedGroup(ZeroOrManyMixin, TypeIncUntypedGroupMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has an untyped subgroup with a data type inc subgroup/dataset/link with quantity '*'
    """
    pass


class TestBuildOneOrManyTypeIncUntypedGroup(OneOrManyMixin, TypeIncUntypedGroupMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has an untyped subgroup with a data type inc subgroup/dataset/link with quantity '+'
    """
    pass


class TestBuildOneTypeIncUntypedGroup(OneMixin, TypeIncUntypedGroupMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has an untyped subgroup with a data type inc subgroup/dataset/link with quantity 1
    """
    pass


class TestBuildTwoTypeIncUntypedGroup(TwoMixin, TypeIncUntypedGroupMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has an untyped subgroup with a data type inc subgroup/dataset/link with quantity 2
    """
    pass


class TestBuildZeroOrOneTypeIncUntypedGroup(ZeroOrOneMixin, TypeIncUntypedGroupMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has an untyped subgroup with a data type inc subgroup/dataset/link with quantity '?'
    """
    pass


# Nested type definition of group/dataset with quantity {'*', '+', 1, 2, '?'}

class TestBuildZeroOrManyTypeDef(ZeroOrManyMixin, TypeDefMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a nested type def with quantity '*'
    """
    pass


class TestBuildOneOrManyTypeDef(OneOrManyMixin, TypeDefMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a nested type def with quantity '+'
    """
    pass


class TestBuildOneTypeDef(OneMixin, TypeDefMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a nested type def with quantity 1
    """
    pass


class TestBuildTwoTypeDef(TwoMixin, TypeDefMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a nested type def with quantity 2
    """
    pass


class TestBuildZeroOrOneTypeDef(ZeroOrOneMixin, TypeDefMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a nested type def with quantity '?'
    """
    pass


# Included groups / included datasets / links with quantity {'*', '+', 1, 2, '?'}

class TestBuildZeroOrManyTypeInc(ZeroOrManyMixin, TypeIncMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a data type inc subgroup/dataset/link with quantity '*'
    """
    pass


class TestBuildOneOrManyTypeInc(OneOrManyMixin, TypeIncMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a data type inc subgroup/dataset/link with quantity '+'
    """
    pass


class TestBuildOneTypeInc(OneMixin, TypeIncMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a data type inc subgroup/dataset/link with quantity 1
    """
    pass


class TestBuildTwoTypeInc(TwoMixin, TypeIncMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a data type inc subgroup/dataset/link with quantity 2
    """
    pass


class TestBuildZeroOrOneTypeInc(ZeroOrOneMixin, TypeIncMixin, BuildQuantityMixin, TestCase):
    """Test building a group that has a data type inc subgroup/dataset/link with quantity '?'
    """
    pass


# Untyped group/dataset with quantity {1, '?'}

class UntypedMixin:

    def setUpManager(self, specs):
        spec_catalog = SpecCatalog()
        schema_file = 'test.yaml'
        for s in specs:
            spec_catalog.register_spec(s, schema_file)
        namespace = SpecNamespace(
            doc='a test namespace',
            name=CORE_NAMESPACE,
            schema=[{'source': schema_file}],
            version='0.1.0',
            catalog=spec_catalog
        )
        namespace_catalog = NamespaceCatalog()
        namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
        type_map = TypeMap(namespace_catalog)
        type_map.register_container_type(CORE_NAMESPACE, 'BasicBucket', BasicBucket)
        self.manager = BuildManager(type_map)

    def create_specs(self, quantity):
        # Type BasicBucket contains:
        # - [quantity] untyped group
        # - [quantity] untyped dataset
        # - [quantity] untyped array dataset
        # quantity can be only '?' or 1
        untyped_group_spec = GroupSpec(
            doc='A test group specification with no data type',
            name='untyped_group',
            quantity=quantity,
        )
        untyped_dataset_spec = DatasetSpec(
            doc='A test dataset specification with no data type',
            name='untyped_dataset',
            dtype='int',
            quantity=quantity,
        )
        untyped_array_dataset_spec = DatasetSpec(
            doc='A test dataset specification with no data type',
            name='untyped_array_dataset',
            dtype='int',
            dims=[None],
            shape=[None],
            quantity=quantity,
        )
        basic_bucket_spec = GroupSpec(
            doc='A test group specification for a data type containing data type',
            name="test_bucket",
            data_type_def='BasicBucket',
            groups=[untyped_group_spec],
            datasets=[untyped_dataset_spec, untyped_array_dataset_spec],
        )
        return [basic_bucket_spec]


class TestBuildOneUntyped(UntypedMixin, TestCase):
    """Test building a group that has an untyped subgroup/dataset with quantity 1.
    """
    def setUp(self):
        specs = self.create_specs(DEF_QUANTITY)
        self.setUpManager(specs)

    def test_build_data(self):
        """Test building a container which contains an untyped empty subgroup and an untyped non-empty dataset."""
        bucket = BasicBucket(name='test_bucket', untyped_dataset=3, untyped_array_dataset=[3])
        # a required untyped empty group builder will be created by default
        untyped_group_builder = GroupBuilder(name='untyped_group')
        untyped_dataset_builder = DatasetBuilder(name='untyped_dataset', data=3)
        untyped_array_dataset_builder = DatasetBuilder(name='untyped_array_dataset', data=[3])
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'untyped_group': untyped_group_builder},
            datasets={'untyped_dataset': untyped_dataset_builder,
                      'untyped_array_dataset': untyped_array_dataset_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'BasicBucket',
                        'object_id': bucket.object_id}
        )
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_empty_data(self):
        """Test building a container which contains an untyped empty subgroup and an untyped empty dataset."""
        bucket = BasicBucket(name='test_bucket')
        # a required untyped empty group builder will be created by default
        untyped_group_builder = GroupBuilder(name='untyped_group')
        # a required untyped empty dataset builder will NOT be created by default
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'untyped_group': untyped_group_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'BasicBucket',
                        'object_id': bucket.object_id}
        )
        msg = "BasicBucket 'test_bucket' is missing required value for attribute 'untyped_dataset'."
        # also raises "BasicBucket 'test_bucket' is missing required value for attribute 'untyped_array_dataset'."
        with self.assertWarnsWith(MissingRequiredBuildWarning, msg):
            builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


class TestBuildZeroOrOneUntyped(UntypedMixin, TestCase):
    """Test building a group that has an untyped subgroup/dataset with quantity '?'.
    """
    def setUp(self):
        specs = self.create_specs(ZERO_OR_ONE)
        self.setUpManager(specs)

    def test_build_data(self):
        """Test building a container which contains an untyped empty subgroup and an untyped non-empty dataset."""
        bucket = BasicBucket(name='test_bucket', untyped_dataset=3, untyped_array_dataset=[3])
        # an optional untyped empty group builder will NOT be created by default
        untyped_dataset_builder = DatasetBuilder(name='untyped_dataset', data=3)
        untyped_array_dataset_builder = DatasetBuilder(name='untyped_array_dataset', data=[3])
        bucket_builder = GroupBuilder(
            name='test_bucket',
            datasets={'untyped_dataset': untyped_dataset_builder,
                      'untyped_array_dataset': untyped_array_dataset_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'BasicBucket',
                        'object_id': bucket.object_id}
        )
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)

    def test_build_empty_data(self):
        """Test building a container which contains an untyped empty subgroup and an untyped empty dataset."""
        bucket = BasicBucket(name='test_bucket')
        # an optional untyped empty group builder will NOT be created by default
        # an optional untyped empty dataset builder will NOT be created by default
        bucket_builder = GroupBuilder(
            name='test_bucket',
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'BasicBucket',
                        'object_id': bucket.object_id}
        )
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)


# Multiple allowed types

class TestBuildMultiTypeInc(BuildQuantityMixin, TestCase):
    """Test build process when a groupspec allows multiple groups/datasets/links with different data types / targets.
    """

    def setUp(self):
        specs = self.create_specs(ZERO_OR_MANY)
        self.setUpManager(specs)

    def create_specs(self, quantity):
        # Type SimpleBucket contains:
        # - [quantity] groups of data_type_inc SimpleFoo and [quantity] group of data_type_inc NotSimpleFoo
        # - [quantity] datasets of data_type_inc SimpleQux and [quantity] datasets of data_type_inc NotSimpleQux
        # - [quantity] links of target_type SimpleFoo and [quantity] links of target_type NotSimpleFoo
        foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='SimpleFoo',
        )
        not_foo_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleFoo',
        )
        qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='SimpleQux',
        )
        not_qux_spec = DatasetSpec(
            doc='A test group specification with a data type',
            data_type_def='NotSimpleQux',
        )
        foo_inc_spec = GroupSpec(
            doc='the SimpleFoos in this bucket',
            data_type_inc='SimpleFoo',
            quantity=quantity
        )
        not_foo_inc_spec = GroupSpec(
            doc='the SimpleFoos in this bucket',
            data_type_inc='NotSimpleFoo',
            quantity=quantity
        )
        qux_inc_spec = DatasetSpec(
            doc='the SimpleQuxs in this bucket',
            data_type_inc='SimpleQux',
            quantity=quantity
        )
        not_qux_inc_spec = DatasetSpec(
            doc='the SimpleQuxs in this bucket',
            data_type_inc='NotSimpleQux',
            quantity=quantity
        )
        foo_link_spec = LinkSpec(
            doc='the links in this bucket',
            target_type='SimpleFoo',
            quantity=quantity
        )
        not_foo_link_spec = LinkSpec(
            doc='the links in this bucket',
            target_type='NotSimpleFoo',
            quantity=quantity
        )
        bucket_spec = GroupSpec(
            doc='A test group specification for a data type containing data type',
            name="test_bucket",
            data_type_def='SimpleBucket',
            groups=[foo_inc_spec, not_foo_inc_spec],
            datasets=[qux_inc_spec, not_qux_inc_spec],
            links=[foo_link_spec, not_foo_link_spec]
        )
        return [foo_spec, not_foo_spec, qux_spec, not_qux_spec, bucket_spec]

    def setUpBucketMapper(self):
        class BucketMapper(ObjectMapper):
            def __init__(self, spec):
                super().__init__(spec)
                self.map_spec('foos', spec.get_data_type('SimpleFoo'))
                self.map_spec('foos', spec.get_data_type('NotSimpleFoo'))
                self.map_spec('quxs', spec.get_data_type('SimpleQux'))
                self.map_spec('quxs', spec.get_data_type('NotSimpleQux'))
                self.map_spec('links', spec.links[0])
                self.map_spec('links', spec.links[1])

        return BucketMapper

    def get_two_bucket_test(self):
        foos = [SimpleFoo('my_foo1'), NotSimpleFoo('my_foo2')]
        quxs = [SimpleQux('my_qux1', data=[1, 2, 3]), NotSimpleQux('my_qux2', data=[4, 5, 6])]
        # NOTE: unlike in the other tests, links cannot map to the same foos in bucket because of a name clash
        links = [SimpleFoo('my_foo3'), NotSimpleFoo('my_foo4')]
        bucket = SimpleBucket(
            name='test_bucket',
            foos=foos,
            quxs=quxs,
            links=links
        )
        foo1_builder = self._create_builder(bucket.foos['my_foo1'])
        foo2_builder = self._create_builder(bucket.foos['my_foo2'])
        foo3_builder = self._create_builder(bucket.links['my_foo3'])
        foo4_builder = self._create_builder(bucket.links['my_foo4'])
        qux1_builder = self._create_builder(bucket.quxs['my_qux1'])
        qux2_builder = self._create_builder(bucket.quxs['my_qux2'])
        foo3_link_builder = LinkBuilder(builder=foo3_builder)
        foo4_link_builder = LinkBuilder(builder=foo4_builder)
        bucket_builder = GroupBuilder(
            name='test_bucket',
            groups={'my_foo1': foo1_builder,
                    'my_foo2': foo2_builder},
            datasets={'my_qux1': qux1_builder,
                      'my_qux2': qux2_builder},
            links={'my_foo3': foo3_link_builder,
                   'my_foo4': foo4_link_builder},
            attributes={'namespace': CORE_NAMESPACE,
                        'data_type': 'SimpleBucket',
                        'object_id': bucket.object_id}
        )
        return bucket, bucket_builder

    def test_build_two(self):
        """Test building a container which contains multiple containers of different types as the spec allows."""
        bucket, bucket_builder = self.get_two_bucket_test()
        builder = self.manager.build(bucket)
        self.assertDictEqual(builder, bucket_builder)



import os

import h5py
import numpy as np
from hdmf.data_utils import DataChunkIterator, DataIO
from hdmf.testing import TestCase
from hdmf.utils import get_data_shape, to_uint_array


class TestGetDataShape(TestCase):

    def test_h5dataset(self):
        """Test get_data_shape on h5py.Datasets of various shapes and maxshape."""
        path = 'test_get_data_shape.h5'
        with h5py.File(path, 'w') as f:
            dset = f.create_dataset('data', data=((1, 2), (3, 4), (5, 6)))
            res = get_data_shape(dset)
            self.assertTupleEqual(res, (3, 2))

            dset = f.create_dataset('shape', shape=(3, 2))
            res = get_data_shape(dset)
            self.assertTupleEqual(res, (3, 2))

            # test that maxshape takes priority
            dset = f.create_dataset('shape_maxshape', shape=(3, 2), maxshape=(None, 100))
            res = get_data_shape(dset)
            self.assertTupleEqual(res, (None, 100))

        os.remove(path)

    def test_dci(self):
        """Test get_data_shape on DataChunkIterators of various shapes and maxshape."""
        dci = DataChunkIterator(dtype=np.dtype(int))
        res = get_data_shape(dci)
        self.assertIsNone(res)

        dci = DataChunkIterator(data=[1, 2])
        res = get_data_shape(dci)
        self.assertTupleEqual(res, (2, ))

        dci = DataChunkIterator(data=[[1, 2], [3, 4], [5, 6]])
        res = get_data_shape(dci)
        self.assertTupleEqual(res, (3, 2))

        # test that maxshape takes priority
        dci = DataChunkIterator(data=[[1, 2], [3, 4], [5, 6]], maxshape=(None, 100))
        res = get_data_shape(dci)
        self.assertTupleEqual(res, (None, 100))

    def test_dataio(self):
        """Test get_data_shape on DataIO of various shapes and maxshape."""
        dio = DataIO(data=[1, 2])
        res = get_data_shape(dio)
        self.assertTupleEqual(res, (2, ))

        dio = DataIO(data=[[1, 2], [3, 4], [5, 6]])
        res = get_data_shape(dio)
        self.assertTupleEqual(res, (3, 2))

        dio = DataIO(data=np.array([[1, 2], [3, 4], [5, 6]]))
        res = get_data_shape(dio)
        self.assertTupleEqual(res, (3, 2))

    def test_list(self):
        """Test get_data_shape on lists of various shapes."""
        res = get_data_shape(list())
        self.assertTupleEqual(res, (0, ))

        res = get_data_shape([1, 2])
        self.assertTupleEqual(res, (2, ))

        res = get_data_shape([[1, 2], [3, 4], [5, 6]])
        self.assertTupleEqual(res, (3, 2))

    def test_tuple(self):
        """Test get_data_shape on tuples of various shapes."""
        res = get_data_shape(tuple())
        self.assertTupleEqual(res, (0, ))

        res = get_data_shape((1, 2))
        self.assertTupleEqual(res, (2, ))

        res = get_data_shape(((1, 2), (3, 4), (5, 6)))
        self.assertTupleEqual(res, (3, 2))

    def test_nparray(self):
        """Test get_data_shape on numpy arrays of various shapes."""
        res = get_data_shape(np.empty([]))
        self.assertTupleEqual(res, tuple())

        res = get_data_shape(np.array([]))
        self.assertTupleEqual(res, (0, ))

        res = get_data_shape(np.array([1, 2]))
        self.assertTupleEqual(res, (2, ))

        res = get_data_shape(np.array([[1, 2], [3, 4], [5, 6]]))
        self.assertTupleEqual(res, (3, 2))

    def test_other(self):
        """Test get_data_shape on miscellaneous edge cases."""
        res = get_data_shape(dict())
        self.assertIsNone(res)

        res = get_data_shape(None)
        self.assertIsNone(res)

        res = get_data_shape([None, None])
        self.assertTupleEqual(res, (2, ))

        res = get_data_shape(object())
        self.assertIsNone(res)

        res = get_data_shape([object(), object()])
        self.assertTupleEqual(res, (2, ))

    def test_string(self):
        """Test get_data_shape on strings and collections of strings."""
        res = get_data_shape('abc')
        self.assertIsNone(res)

        res = get_data_shape(('a', 'b'))
        self.assertTupleEqual(res, (2, ))

        res = get_data_shape((('a', 'b'), ('c', 'd'), ('e', 'f')))
        self.assertTupleEqual(res, (3, 2))

    def test_set(self):
        """Test get_data_shape on sets, which have __len__ but are not subscriptable."""
        res = get_data_shape(set())
        self.assertTupleEqual(res, (0, ))

        res = get_data_shape({1, 2})
        self.assertTupleEqual(res, (2, ))

    def test_arbitrary_iterable_with_len(self):
        """Test get_data_shape with strict_no_data_load=True on an arbitrary iterable object with __len__."""

        class MyIterable:
            """Iterable class without shape or maxshape, where loading the first element raises an error."""

            def __len__(self):
                return 10

            def __iter__(self):
                return self

            def __next__(self):
                raise DataLoadedError()

        class DataLoadedError(Exception):
            pass

        data = MyIterable()
        with self.assertRaises(DataLoadedError):
            get_data_shape(data)  # test that data is loaded

        res = get_data_shape(data, strict_no_data_load=True)  # no error raised means data was not loaded
        self.assertIsNone(res)

    def test_strict_no_data_load(self):
        """Test get_data_shape with strict_no_data_load=True on nested lists/tuples is the same as when it is False."""
        res = get_data_shape([[1, 2], [3, 4], [5, 6]], strict_no_data_load=True)
        self.assertTupleEqual(res, (3, 2))

        res = get_data_shape(((1, 2), (3, 4), (5, 6)), strict_no_data_load=True)
        self.assertTupleEqual(res, (3, 2))


class TestToUintArray(TestCase):

    def test_ndarray_uint(self):
        arr = np.array([0, 1, 2], dtype=np.uint32)
        res = to_uint_array(arr)
        np.testing.assert_array_equal(res, arr)

    def test_ndarray_int(self):
        arr = np.array([0, 1, 2], dtype=np.int32)
        res = to_uint_array(arr)
        np.testing.assert_array_equal(res, arr)

    def test_ndarray_int_neg(self):
        arr = np.array([0, -1, 2], dtype=np.int32)
        with self.assertRaisesWith(ValueError, 'Cannot convert negative integer values to uint.'):
            to_uint_array(arr)

    def test_ndarray_float(self):
        arr = np.array([0, 1, 2], dtype=np.float64)
        with self.assertRaisesWith(ValueError, 'Cannot convert array of dtype float64 to uint.'):
            to_uint_array(arr)

    def test_list_int(self):
        arr = [0, 1, 2]
        res = to_uint_array(arr)
        expected = np.array([0, 1, 2], dtype=np.uint32)
        np.testing.assert_array_equal(res, expected)

    def test_list_int_neg(self):
        arr = [0, -1, 2]
        with self.assertRaisesWith(ValueError, 'Cannot convert negative integer values to uint.'):
            to_uint_array(arr)

    def test_list_float(self):
        arr = [0., 1., 2.]
        with self.assertRaisesWith(ValueError, 'Cannot convert array of dtype float64 to uint.'):
            to_uint_array(arr)



from h5py import File

from hdmf.backends.hdf5 import HDF5IO
from hdmf.common import Container, get_manager, get_hdf5io
from hdmf.spec import NamespaceCatalog
from hdmf.testing import TestCase, remove_test_file

from tests.unit.utils import get_temp_filepath


class TestCacheSpec(TestCase):
    """Test caching spec specifically with the namespaces provided by hdmf.common.

    See also TestCacheSpec in tests/unit/test_io_hdf5_h5tools.py.
    """

    def setUp(self):
        self.manager = get_manager()
        self.path = get_temp_filepath()
        self.container = Container('dummy')

    def tearDown(self):
        remove_test_file(self.path)

    def test_write_no_cache_spec(self):
        """Roundtrip test for not writing spec."""
        with HDF5IO(self.path, manager=self.manager, mode="a") as io:
            io.write(self.container, cache_spec=False)
        with File(self.path, 'r') as f:
            self.assertNotIn('specifications', f)

    def test_write_cache_spec(self):
        """Roundtrip test for writing spec and reading it back in."""
        with HDF5IO(self.path, manager=self.manager, mode="a") as io:
            io.write(self.container)
        with File(self.path, 'r') as f:
            self.assertIn('specifications', f)
        self._check_spec()

    def test_write_cache_spec_injected(self):
        """Roundtrip test for writing spec and reading it back in when HDF5IO is passed an open h5py.File."""
        with File(self.path, 'w') as fil:
            with HDF5IO(self.path, manager=self.manager, file=fil, mode='a') as io:
                io.write(self.container)
        with File(self.path, 'r') as f:
            self.assertIn('specifications', f)
        self._check_spec()

    def _check_spec(self):
        ns_catalog = NamespaceCatalog()
        HDF5IO.load_namespaces(ns_catalog, self.path)
        self.maxDiff = None
        for namespace in self.manager.namespace_catalog.namespaces:
            with self.subTest(namespace=namespace):
                original_ns = self.manager.namespace_catalog.get_namespace(namespace)
                cached_ns = ns_catalog.get_namespace(namespace)
                ns_fields_to_check = list(original_ns.keys())
                ns_fields_to_check.remove('schema')  # schema fields will not match, so reset
                for ns_field in ns_fields_to_check:
                    with self.subTest(namespace_field=ns_field):
                        self.assertEqual(original_ns[ns_field], cached_ns[ns_field])
                for dt in original_ns.get_registered_types():
                    with self.subTest(data_type=dt):
                        original_spec = original_ns.get_spec(dt)
                        cached_spec = cached_ns.get_spec(dt)
                        with self.subTest('Data type spec is read back in'):
                            self.assertIsNotNone(cached_spec)
                        with self.subTest('Cached spec matches original spec'):
                            self.assertDictEqual(original_spec, cached_spec)


class TestGetHdf5IO(TestCase):

    def setUp(self):
        self.path = get_temp_filepath()

    def tearDown(self):
        remove_test_file(self.path)

    def test_gethdf5io(self):
        """Test the get_hdf5io convenience method with manager=None."""
        with get_hdf5io(self.path, "w") as io:
            self.assertIsNotNone(io.manager)

    def test_gethdf5io_manager(self):
        """Test the get_hdf5io convenience method with manager set."""
        manager = get_manager()
        with get_hdf5io(self.path, "w", manager=manager) as io:
            self.assertIs(io.manager, manager)



import pandas as pd
from hdmf.common import DynamicTable
from hdmf.common.resources import ExternalResources, Key, Resource
from hdmf import Data
from hdmf.testing import TestCase, H5RoundTripMixin
import numpy as np
import unittest
from tests.unit.build_tests.test_io_map import Bar
from tests.unit.utils import create_test_type_map, CORE_NAMESPACE
from hdmf.spec import GroupSpec, AttributeSpec, DatasetSpec


class TestExternalResources(H5RoundTripMixin, TestCase):

    def setUpContainer(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        er.add_ref(
            container='uuid2', key='key2',
            resource_name='resource21', resource_uri='resource_uri21', entity_id="id12", entity_uri='url21')
        return er

    @unittest.skip('Outdated do to privatization')
    def test_piecewise_add(self):
        er = ExternalResources(name='terms')

        # this is the term the user wants to use. They will need to specify this
        key = er._add_key('mouse')

        resource1 = er._add_resource(resource='resource0', uri='resource_uri0')
        # the user will have to supply this info as well. This is the information
        # needed to retrieve info about the controled term
        er._add_entity(key, resource1, '10090', 'uri')

        # The user can also pass in the container or it can be wrapped up under NWBFILE
        obj = er._add_object('object', 'species')

        # This could also be wrapped up under NWBFile
        er._add_object_key(obj, key)

        self.assertEqual(er.keys.data, [('mouse',)])
        self.assertEqual(er.entities.data,
                         [(0, 0, '10090', 'uri')])
        self.assertEqual(er.objects.data, [('object', 'species')])

    def test_to_dataframe(self):
        # Setup complex external resources with keys reused across objects and
        # multiple resources per key
        er = ExternalResources(name='example')
        # Add a species dataset with 2 keys
        data1 = Data(
            name='data_name',
            data=np.array(
                [('Mus musculus', 9, 81.0), ('Homo sapiens', 3, 27.0)],
                dtype=[('species', 'U14'), ('age', 'i4'), ('weight', 'f4')]
            )
        )
        k1, r1, e1 = er.add_ref(
            container=data1,
            field='species',
            key='Mus musculus',
            resource_name='NCBI_Taxonomy',
            resource_uri='https://www.ncbi.nlm.nih.gov/taxonomy',
            entity_id='NCBI:txid10090',
            entity_uri='https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090'
        )
        k2, r2, e2 = er.add_ref(
            container=data1,
            field='species',
            key='Homo sapiens',
            resource_name='NCBI_Taxonomy',
            resource_uri='https://www.ncbi.nlm.nih.gov/taxonomy',
            entity_id='NCBI:txid9606',
            entity_uri='https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9606'
        )
        # Add a second species dataset that uses the same keys as the first dataset and add an additional key
        data2 = Data(name="species", data=['Homo sapiens', 'Mus musculus', 'Pongo abelii'])
        o2 = er._add_object(data2, relative_path='', field='')
        er._add_object_key(o2, k1)
        er._add_object_key(o2, k2)
        k2, r2, e2 = er.add_ref(
            container=data2,
            field='',
            key='Pongo abelii',
            resource_name='NCBI_Taxonomy',
            resource_uri='https://www.ncbi.nlm.nih.gov/taxonomy',
            entity_id='NCBI:txid9601',
            entity_uri='https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9601'
        )
        # Add a third data object, this time with 2 entities for a key
        data3 = Data(name="genotypes", data=['Rorb'])
        k3, r3, e3 = er.add_ref(
            container=data3,
            field='',
            key='Rorb',
            resource_name='MGI Database',
            resource_uri='http://www.informatics.jax.org/',
            entity_id='MGI:1346434',
            entity_uri='http://www.informatics.jax.org/marker/MGI:1343464'
        )
        _ = er.add_ref(
            container=data3,
            field='',
            key=k3,
            resource_name='Ensembl',
            resource_uri='https://uswest.ensembl.org/index.html',
            entity_id='ENSG00000198963',
            entity_uri='https://uswest.ensembl.org/Homo_sapiens/Gene/Summary?db=core;g=ENSG00000198963'
        )
        # Convert to dataframe and compare against the expected result
        result_df = er.to_dataframe()
        expected_df_data = \
            {'objects_idx': {0: 0, 1: 0, 2: 1, 3: 1, 4: 1, 5: 2, 6: 2},
             'object_id': {0: data1.object_id, 1: data1.object_id,
                           2: data2.object_id, 3: data2.object_id, 4: data2.object_id,
                           5: data3.object_id, 6: data3.object_id},
             'relative_path': {0: '', 1: '', 2: '', 3: '', 4: '', 5: '', 6: ''},
             'field': {0: 'species', 1: 'species', 2: '', 3: '', 4: '', 5: '', 6: ''},
             'keys_idx': {0: 0, 1: 1, 2: 0, 3: 1, 4: 2, 5: 3, 6: 3},
             'key': {0: 'Mus musculus', 1: 'Homo sapiens', 2: 'Mus musculus', 3: 'Homo sapiens',
                     4: 'Pongo abelii', 5: 'Rorb', 6: 'Rorb'},
             'resources_idx': {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, 5: 1, 6: 2},
             'resource': {0: 'NCBI_Taxonomy', 1: 'NCBI_Taxonomy', 2: 'NCBI_Taxonomy', 3: 'NCBI_Taxonomy',
                          4: 'NCBI_Taxonomy', 5: 'MGI Database', 6: 'Ensembl'},
             'resource_uri': {0: 'https://www.ncbi.nlm.nih.gov/taxonomy', 1: 'https://www.ncbi.nlm.nih.gov/taxonomy',
                              2: 'https://www.ncbi.nlm.nih.gov/taxonomy', 3: 'https://www.ncbi.nlm.nih.gov/taxonomy',
                              4: 'https://www.ncbi.nlm.nih.gov/taxonomy', 5: 'http://www.informatics.jax.org/',
                              6: 'https://uswest.ensembl.org/index.html'},
             'entities_idx': {0: 0, 1: 1, 2: 0, 3: 1, 4: 2, 5: 3, 6: 4},
             'entity_id': {0: 'NCBI:txid10090', 1: 'NCBI:txid9606', 2: 'NCBI:txid10090', 3: 'NCBI:txid9606',
                           4: 'NCBI:txid9601', 5: 'MGI:1346434', 6: 'ENSG00000198963'},
             'entity_uri': {0: 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090',
                            1: 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9606',
                            2: 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090',
                            3: 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9606',
                            4: 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9601',
                            5: 'http://www.informatics.jax.org/marker/MGI:1343464',
                            6: 'https://uswest.ensembl.org/Homo_sapiens/Gene/Summary?db=core;g=ENSG00000198963'}}
        expected_df = pd.DataFrame.from_dict(expected_df_data)
        pd.testing.assert_frame_equal(result_df, expected_df)

        # Convert to dataframe with categories and compare against the expected result
        result_df = er.to_dataframe(use_categories=True)
        cols_with_categories = [
            ('objects', 'objects_idx'), ('objects', 'object_id'), ('objects', 'relative_path'), ('objects', 'field'),
            ('keys', 'keys_idx'), ('keys', 'key'),
            ('resources', 'resources_idx'), ('resources', 'resource'), ('resources', 'resource_uri'),
            ('entities', 'entities_idx'), ('entities', 'entity_id'), ('entities', 'entity_uri')]
        expected_df_data = {c: expected_df_data[c[1]] for c in cols_with_categories}
        expected_df = pd.DataFrame.from_dict(expected_df_data)
        pd.testing.assert_frame_equal(result_df, expected_df)

    def test_assert_external_resources_equal(self):
        er_left = ExternalResources('terms')
        er_left.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        er_right = ExternalResources('terms')
        er_right.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        self.assertTrue(ExternalResources.assert_external_resources_equal(er_left,
                                                                          er_right))

    def test_invalid_keys_assert_external_resources_equal(self):
        er_left = ExternalResources('terms')
        er_left.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        er_right = ExternalResources('terms')
        er_right.add_ref(
            container='invalid', key='invalid',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        with self.assertRaises(AssertionError):
            ExternalResources.assert_external_resources_equal(er_left,
                                                              er_right)

    def test_invalid_objects_assert_external_resources_equal(self):
        er_left = ExternalResources('terms')
        er_left.add_ref(
            container='invalid', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        er_right = ExternalResources('terms')
        er_right.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        with self.assertRaises(AssertionError):
            ExternalResources.assert_external_resources_equal(er_left,
                                                              er_right)

    def test_invalid_resources_assert_external_resources_equal(self):
        er_left = ExternalResources('terms')
        er_left.add_ref(
            container='uuid1', key='key1',
            resource_name='invalid', resource_uri='invalid',
            entity_id="id11", entity_uri='url11')

        er_right = ExternalResources('terms')
        er_right.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        with self.assertRaises(AssertionError):
            ExternalResources.assert_external_resources_equal(er_left,
                                                              er_right)

    def test_invalid_entity_assert_external_resources_equal(self):
        er_left = ExternalResources('terms')
        er_left.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="invalid", entity_uri='invalid')

        er_right = ExternalResources('terms')
        er_right.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        with self.assertRaises(AssertionError):
            ExternalResources.assert_external_resources_equal(er_left,
                                                              er_right)

    def test_invalid_object_keys_assert_external_resources_equal(self):
        er_left = ExternalResources('terms')
        er_left.add_ref(
            container='invalid', key='invalid',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        er_right = ExternalResources('terms')
        er_right._add_key('key')
        er_right.add_ref(
            container='uuid1', key='key1',
            resource_name='resource11', resource_uri='resource_uri11',
            entity_id="id11", entity_uri='url11')

        with self.assertRaises(AssertionError):
            ExternalResources.assert_external_resources_equal(er_left,
                                                              er_right)

    def test_add_ref(self):
        er = ExternalResources(name='terms')
        data = Data(name="species", data=['Homo sapiens', 'Mus musculus'])
        er.add_ref(
            container=data, key='key1',
            resource_name='resource1', resource_uri='uri1',
            entity_id='entity_id1', entity_uri='entity1')
        self.assertEqual(er.keys.data, [('key1',)])
        self.assertEqual(er.resources.data, [('resource1', 'uri1')])
        self.assertEqual(er.entities.data, [(0, 0, 'entity_id1', 'entity1')])
        self.assertEqual(er.objects.data, [(data.object_id, '', '')])

    def test_to_tsv_and_from_tsv(self):
        # write er to file
        self.container.to_tsv(path=self.export_filename)
        # read er back from file and compare
        er_obj = ExternalResources.from_tsv(path=self.export_filename)
        # Check that the data is correct
        ExternalResources.assert_external_resources_equal(er_obj, self.container, check_dtype=False)

    def test_to_tsv_and_from_tsv_missing_keyidx(self):
        # write er to file
        df = self.container.to_dataframe(use_categories=True)
        df.at[0, ('keys', 'keys_idx')] = 10  # Change key_ix 0 to 10
        df.to_csv(self.export_filename, sep='\t')
        # read er back from file and compare
        msg = "Missing keys_idx entries [0, 2, 3, 4, 5, 6, 7, 8, 9]"
        with self.assertRaisesWith(ValueError, msg):
            _ = ExternalResources.from_tsv(path=self.export_filename)

    def test_to_tsv_and_from_tsv_missing_objectidx(self):
        # write er to file
        df = self.container.to_dataframe(use_categories=True)
        df.at[0, ('objects', 'objects_idx')] = 10  # Change key_ix 0 to 10
        df.to_csv(self.export_filename, sep='\t')
        # read er back from file and compare
        msg = "Missing objects_idx entries [0, 2, 3, 4, 5, 6, 7, 8, 9]"
        with self.assertRaisesWith(ValueError, msg):
            _ = ExternalResources.from_tsv(path=self.export_filename)

    def test_to_tsv_and_from_tsv_missing_resourcesidx(self):
        # write er to file
        df = self.container.to_dataframe(use_categories=True)
        df.at[0, ('resources', 'resources_idx')] = 10  # Change key_ix 0 to 10
        df.to_csv(self.export_filename, sep='\t')
        # read er back from file and compare
        msg = "Missing resources_idx entries [0, 2, 3, 4, 5, 6, 7, 8, 9]"
        with self.assertRaisesWith(ValueError, msg):
            _ = ExternalResources.from_tsv(path=self.export_filename)

    def test_to_tsv_and_from_tsv_missing_entitiesidx(self):
        # write er to file
        df = self.container.to_dataframe(use_categories=True)
        df.at[0, ('entities', 'entities_idx')] = 10  # Change key_ix 0 to 10
        df.to_csv(self.export_filename, sep='\t')
        # read er back from file and compare
        msg = "Missing entities_idx entries [0, 2, 3, 4, 5, 6, 7, 8, 9]"
        with self.assertRaisesWith(ValueError, msg):
            _ = ExternalResources.from_tsv(path=self.export_filename)

    def test_add_ref_duplicate_resource(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1',
            resource_name='resource0', resource_uri='uri0',
            entity_id='entity_id1', entity_uri='entity1')
        er.add_ref(
            container='uuid2', key='key2',
            resource_name='resource0', resource_uri='uri0',
            entity_id='entity_id2', entity_uri='entity2')
        resource_list = er.resources.which(resource='resource0')
        self.assertEqual(len(resource_list), 1)

    def test_add_ref_bad_arg(self):
        er = ExternalResources(name='terms')
        resource1 = er._add_resource(resource='resource0', uri='resource_uri0')
        # The contents of the message are not important. Just make sure an error is raised
        with self.assertRaises(ValueError):
            er.add_ref(
                'uuid1', key='key1', resource_name='resource1',
                resource_uri='uri1', entity_id='resource_id1')
        with self.assertRaises(ValueError):
            er.add_ref('uuid1', key='key1', resource_name='resource1', resource_uri='uri1', entity_uri='uri1')
        with self.assertRaises(ValueError):
            er.add_ref('uuid1', key='key1', resource_name='resource1', resource_uri='uri1')
        with self.assertRaises(TypeError):
            er.add_ref('uuid1')
        with self.assertRaises(ValueError):
            er.add_ref('uuid1', key='key1', resource_name='resource1')
        with self.assertRaises(ValueError):
            er.add_ref(
                'uuid1', key='key1', resources_idx=resource1,
                resource_name='resource1', resource_uri='uri1')

    def test_add_ref_two_resources(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        er.add_ref(
            container='uuid1', key=er.get_key(key_name='key1'), resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')
        self.assertEqual(er.keys.data, [('key1',)])
        self.assertEqual(er.resources.data,
                         [('resource1',  'resource_uri1'),
                          ('resource2', 'resource_uri2')])
        self.assertEqual(er.objects.data, [('uuid1', '', '')])
        self.assertEqual(er.entities.data, [(0, 0, 'id11', 'url11'), (0, 1, 'id12', 'url21')])

    def test_get_resources(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        resource = er.get_resource('resource1')
        self.assertIsInstance(resource, Resource)
        with self.assertRaises(ValueError):
            er.get_resource('unknown_resource')

    def test_add_ref_two_keys(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        er.add_ref(
            container='uuid2', key='key2', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')

        self.assertEqual(er.keys.data, [('key1',), ('key2',)])
        self.assertEqual(er.resources.data,
                         [('resource1',  'resource_uri1'),
                          ('resource2', 'resource_uri2')])
        self.assertEqual(er.entities.data, [(0, 0, 'id11', 'url11'), (1, 1, 'id12', 'url21')])

        self.assertEqual(er.objects.data, [('uuid1', '', ''),
                                           ('uuid2', '', '')])

    def test_add_ref_same_key_diff_objfield(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        er.add_ref(
            container='uuid2', key='key1', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')

        self.assertEqual(er.keys.data, [('key1',), ('key1',)])
        self.assertEqual(er.entities.data, [(0, 0, 'id11', 'url11'), (1, 1, 'id12', 'url21')])
        self.assertEqual(er.resources.data,
                         [('resource1',  'resource_uri1'),
                          ('resource2', 'resource_uri2')])
        self.assertEqual(er.objects.data, [('uuid1', '', ''),
                                           ('uuid2', '', '')])

    def test_add_ref_same_keyname(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        er.add_ref(
            container='uuid2', key='key1', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')
        er.add_ref(
            container='uuid3', key='key1', resource_name='resource3',
            resource_uri='resource_uri3', entity_id="id13", entity_uri='url31')

        self.assertEqual(er.keys.data, [('key1',), ('key1',), ('key1',)])
        self.assertEqual(er.resources.data,
                         [('resource1',  'resource_uri1'),
                          ('resource2', 'resource_uri2'),
                          ('resource3', 'resource_uri3')])
        self.assertEqual(
            er.entities.data,
            [(0, 0, 'id11', 'url11'),
             (1, 1, 'id12', 'url21'),
             (2, 2, 'id13', 'url31')])
        self.assertEqual(er.objects.data, [('uuid1', '', ''),
                                           ('uuid2', '', ''),
                                           ('uuid3', '', '')])

    def test_get_keys(self):
        er = ExternalResources(name='terms')

        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        er.add_ref(
            container='uuid2', key='key2', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')
        er.add_ref(
            container='uuid1', key=er.get_key(key_name='key1'), resource_name='resource3',
            resource_uri='resource_uri3', entity_id="id13", entity_uri='url31')
        received = er.get_keys()

        expected = pd.DataFrame(
            data=[['key1', 0, 'id11', 'url11'],
                  ['key1', 2, 'id13', 'url31'],
                  ['key2', 1, 'id12', 'url21']],
            columns=['key_name', 'resources_idx', 'entity_id', 'entity_uri'])
        pd.testing.assert_frame_equal(received, expected)

    def test_get_keys_subset(self):
        er = ExternalResources(name='terms')
        er.add_ref(
            container='uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        er.add_ref(
            container='uuid2', key='key2', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')
        er.add_ref(
            container='uuid1', key=er.get_key(key_name='key1'), resource_name='resource3',
            resource_uri='resource_uri3', entity_id="id13", entity_uri='url31')
        key = er.keys.row[0]
        received = er.get_keys(keys=key)

        expected = pd.DataFrame(
            data=[['key1', 0, 'id11', 'url11'],
                  ['key1', 2, 'id13', 'url31']],
            columns=['key_name', 'resources_idx', 'entity_id', 'entity_uri'])
        pd.testing.assert_frame_equal(received, expected)

    def test_get_object_resources(self):
        er = ExternalResources(name='terms')
        table = DynamicTable(name='test_table', description='test table description')
        table.add_column(name='test_col', description='test column description')
        table.add_row(test_col='Mouse')

        er.add_ref(container=table, attribute='test_col', key='Mouse',
                   resource_name='NCBI_Taxonomy',
                   resource_uri='https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi',
                   entity_id='NCBI:txid10090',
                   entity_uri='https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090',
                   )
        received = er.get_object_resources(table['test_col'])
        expected = pd.DataFrame(
            data=[[0, 0, 'NCBI:txid10090', 'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090']],
            columns=['keys_idx', 'resource_idx', 'entity_id', 'entity_uri'])
        pd.testing.assert_frame_equal(received, expected)

    def test_object_key_unqiueness(self):
        er = ExternalResources(name='terms')
        data = Data(name='data_name', data=np.array([('Mus musculus', 9, 81.0), ('Homo sapien', 3, 27.0)],
                    dtype=[('species', 'U14'), ('age', 'i4'), ('weight', 'f4')]))

        er.add_ref(container=data, key='Mus musculus', resource_name='NCBI_Taxonomy',
                   resource_uri='https://www.ncbi.nlm.nih.gov/taxonomy',
                   entity_id='NCBI:txid10090',
                   entity_uri='https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090')
        existing_key = er.get_key('Mus musculus')
        er.add_ref(container=data, key=existing_key, resource_name='resource2',
                   resource_uri='resource_uri2',
                   entity_id='entity2',
                   entity_uri='entity_uri2')

        self.assertEqual(er.object_keys.data, [(0, 0)])

    def test_check_object_field_add(self):
        er = ExternalResources(name='terms')
        data = Data(name="species", data=['Homo sapiens', 'Mus musculus'])
        er._check_object_field('uuid1', '')
        er._check_object_field(data, '')

        self.assertEqual(er.objects.data, [('uuid1', '', ''), (data.object_id, '', '')])

    def test_check_object_field_multi_error(self):
        er = ExternalResources(name='terms')
        data = Data(name="species", data=['Homo sapiens', 'Mus musculus'])
        er._check_object_field(data, '')
        er._add_object(data, '', '')
        with self.assertRaises(ValueError):
            er._check_object_field(data, '')

    def test_check_object_field_not_in_obj_table(self):
        er = ExternalResources(name='terms')
        data = Data(name="species", data=['Homo sapiens', 'Mus musculus'])
        with self.assertRaises(ValueError):
            er._check_object_field(container=data, relative_path='', field='', create=False)

    def test_add_ref_attribute(self):
        # Test to make sure the attribute object is being used for the id
        # for the external reference.
        table = DynamicTable(name='table', description='table')
        table.add_column(name='col1', description="column")
        table.add_row(id=0, col1='data')

        er = ExternalResources(name='example')
        er.add_ref(container=table,
                   attribute='id',
                   key='key1',
                   resource_name='resource0',
                   resource_uri='resource0_uri',
                   entity_id='entity_0',
                   entity_uri='entity_0_uri')

        self.assertEqual(er.keys.data, [('key1',)])
        self.assertEqual(er.resources.data, [('resource0', 'resource0_uri')])
        self.assertEqual(er.entities.data, [(0, 0, 'entity_0', 'entity_0_uri')])
        self.assertEqual(er.objects.data, [(table.id.object_id, '', '')])

    def test_add_ref_column_as_attribute(self):
        # Test to make sure the attribute object is being used for the id
        # for the external reference.
        table = DynamicTable(name='table', description='table')
        table.add_column(name='col1', description="column")
        table.add_row(id=0, col1='data')

        er = ExternalResources(name='example')
        er.add_ref(container=table,
                   attribute='col1',
                   key='key1',
                   resource_name='resource0',
                   resource_uri='resource0_uri',
                   entity_id='entity_0',
                   entity_uri='entity_0_uri')

        self.assertEqual(er.keys.data, [('key1',)])
        self.assertEqual(er.resources.data, [('resource0', 'resource0_uri')])
        self.assertEqual(er.entities.data, [(0, 0, 'entity_0', 'entity_0_uri')])
        self.assertEqual(er.objects.data, [(table['col1'].object_id, '', '')])

    def test_add_ref_compound_data(self):
        er = ExternalResources(name='example')

        data = Data(
            name='data_name',
            data=np.array(
                [('Mus musculus', 9, 81.0), ('Homo sapiens', 3, 27.0)],
                dtype=[('species', 'U14'), ('age', 'i4'), ('weight', 'f4')]))
        er.add_ref(
            container=data,
            field='species',
            key='Mus musculus',
            resource_name='NCBI_Taxonomy',
            resource_uri='resource0_uri',
            entity_id='NCBI:txid10090',
            entity_uri='entity_0_uri'
        )
        self.assertEqual(er.keys.data, [('Mus musculus',)])
        self.assertEqual(er.resources.data, [('NCBI_Taxonomy', 'resource0_uri')])
        self.assertEqual(er.entities.data, [(0, 0, 'NCBI:txid10090', 'entity_0_uri')])
        self.assertEqual(er.objects.data, [(data.object_id, '', 'species')])


class TestExternalResourcesNestedAttributes(TestCase):

    def setUp(self):
        self.attr1 = AttributeSpec(name='attr1', doc='a string attribute', dtype='text')
        self.attr2 = AttributeSpec(name='attr2', doc='an integer attribute', dtype='int')
        self.attr3 = AttributeSpec(name='attr3', doc='an integer attribute', dtype='int')
        self.bar_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Bar',
            datasets=[
                DatasetSpec(
                    doc='a dataset',
                    dtype='int',
                    name='data',
                    attributes=[self.attr2]
                )
            ],
            attributes=[self.attr1])

        specs = [self.bar_spec]
        containers = {'Bar': Bar}
        self.type_map = create_test_type_map(specs, containers)
        self.spec_catalog = self.type_map.namespace_catalog.get_namespace(CORE_NAMESPACE).catalog

        self.cls = self.type_map.get_dt_container_cls(self.bar_spec.data_type)
        self.bar = self.cls(name='bar', data=[1], attr1='attr1', attr2=1)
        obj_mapper_bar = self.type_map.get_map(self.bar)
        obj_mapper_bar.map_spec('attr2', spec=self.attr2)

    def test_add_ref_nested(self):
        table = DynamicTable(name='table', description='table')
        table.add_column(name='col1', description="column")
        table.add_row(id=0, col1='data')

        er = ExternalResources(name='example')
        er.add_ref(container=table,
                   attribute='description',
                   key='key1',
                   resource_name='resource0',
                   resource_uri='resource0_uri',
                   entity_id='entity_0',
                   entity_uri='entity_0_uri')
        self.assertEqual(er.keys.data, [('key1',)])
        self.assertEqual(er.resources.data, [('resource0', 'resource0_uri')])
        self.assertEqual(er.entities.data, [(0, 0, 'entity_0', 'entity_0_uri')])
        self.assertEqual(er.objects.data, [(table.object_id, 'DynamicTable/description', '')])

    def test_add_ref_deep_nested(self):
        er = ExternalResources(name='example', type_map=self.type_map)
        er.add_ref(container=self.bar,
                   attribute='attr2',
                   key='key1',
                   resource_name='resource0',
                   resource_uri='resource0_uri',
                   entity_id='entity_0',
                   entity_uri='entity_0_uri')
        self.assertEqual(er.objects.data[0][1], 'Bar/data/attr2', '')


class TestExternalResourcesGetKey(TestCase):

    def setUp(self):
        self.er = ExternalResources(name='terms')

    def test_get_key(self):
        self.er.add_ref(
            'uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        self.er.add_ref(
            'uuid2', key='key1', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')

        keys = self.er.get_key('key1', 'uuid2', '')
        self.assertIsInstance(keys, Key)
        self.assertEqual(keys.idx, 1)

    def test_get_key_bad_arg(self):
        self.er._add_key('key2')
        self.er.add_ref(
            'uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        with self.assertRaises(ValueError):
            self.er.get_key('key2', 'uuid1', '')

    @unittest.skip('Outdated do to privatization')
    def test_get_key_without_container(self):
        self.er = ExternalResources(name='terms')
        self.er._add_key('key1')
        keys = self.er.get_key('key1')
        self.assertIsInstance(keys, Key)

    def test_get_key_w_object_info(self):
        self.er.add_ref(
            'uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        self.er.add_ref(
            'uuid2', key='key1', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')
        keys = self.er.get_key('key1', 'uuid1', '')
        self.assertIsInstance(keys, Key)
        self.assertEqual(keys.key, 'key1')

    def test_get_key_w_bad_object_info(self):
        self.er.add_ref(
            'uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        self.er.add_ref(
            'uuid2', key='key1', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')

        with self.assertRaisesRegex(ValueError, "No key 'key2'"):
            self.er.get_key('key2', 'uuid1', '')

    def test_get_key_doesnt_exist(self):
        self.er.add_ref(
            'uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        self.er.add_ref(
            'uuid2', key='key1', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url21')
        with self.assertRaisesRegex(ValueError, "key 'bad_key' does not exist"):
            self.er.get_key('bad_key')

    @unittest.skip('Outdated do to privatization')
    def test_get_key_same_keyname_all(self):
        self.er = ExternalResources(name='terms')
        key1 = self.er._add_key('key1')
        key2 = self.er._add_key('key1')
        self.er.add_ref(
            'uuid1', key=key1, resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        self.er.add_ref(
            'uuid2', key=key2, resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url12')
        self.er.add_ref(
            'uuid1', key=self.er.get_key('key1', 'uuid1', ''), resource_name='resource3',
            resource_uri='resource_uri3', entity_id="id13", entity_uri='url13')

        keys = self.er.get_key('key1')

        self.assertIsInstance(keys, Key)
        self.assertEqual(keys[0].key, 'key1')
        self.assertEqual(keys[1].key, 'key1')

    def test_get_key_same_keyname_specific(self):
        self.er = ExternalResources(name='terms')

        self.er.add_ref(
            'uuid1', key='key1', resource_name='resource1',
            resource_uri='resource_uri1', entity_id="id11", entity_uri='url11')
        self.er.add_ref(
            'uuid2', key='key2', resource_name='resource2',
            resource_uri='resource_uri2', entity_id="id12", entity_uri='url12')
        self.er.add_ref(
            'uuid1', key=self.er.get_key('key1', 'uuid1', ''), resource_name='resource3',
            resource_uri='resource_uri3', entity_id="id13", entity_uri='url13')

        keys = self.er.get_key('key1', 'uuid1', '')
        self.assertIsInstance(keys, Key)
        self.assertEqual(keys.key, 'key1')
        self.assertEqual(self.er.keys.data, [('key1',), ('key2',)])



from hdmf.common import SimpleMultiContainer
from hdmf.container import Container, Data
from hdmf.testing import TestCase, H5RoundTripMixin


class SimpleMultiContainerRoundTrip(H5RoundTripMixin, TestCase):

    def setUpContainer(self):
        containers = [
            Container('container1'),
            Container('container2'),
            Data('data1', [0, 1, 2, 3, 4]),
            Data('data2', [0.0, 1.0, 2.0, 3.0, 4.0]),
        ]
        multi_container = SimpleMultiContainer(name='multi', containers=containers)
        return multi_container
